package Moab::Log::Log4perl;
use strict;
use warnings;
use Cwd qw( abs_path );
use Storable qw(dclone);
use Parse::RecDescent;
use Data::Dumper;
use Log::Log4perl qw(:easy);
use File::Basename;
use FindBin qw($Bin);

my %defaultAppenderAttributes = (
    permissions   => 666,
    maxsize       => 10000000,
    rolloverlimit => 7,
    loglevel      => "info",
    layout        => "%d{yyyy-MM-dd	HH:mm:ss.SSS} %L  %-5p	%m%n"
);

Log::Log4perl->easy_init(
    {
        level  => $DEBUG,
        file   => ">>/dev/null",
        layout => $defaultAppenderAttributes{"layout"}
    }
);

sub new
{
    my $nameSpace = shift;
    my %args      = @_;
    _init(\%args);
    return bless \%args, $nameSpace;
}

sub _init
{
    my $args = shift;
    my %appenders;
    $args->{"logger"} = get_logger();
    eval {
        %appenders = map { lc($_) => {} } @{$args->{"appenders"}};
    };
    if ($@)
    {
        $args->{"logger"}->logcroak("appenders not array reference\n");
    }
    $args->{"appenders"} = \%appenders;
    my $defaultLogName = __PACKAGE__;
    $defaultLogName =~ s/::/_/g;
    $args->{"default_appender_values"} = dclone(\%defaultAppenderAttributes);
    if (defined($args->{"config_file"}))
    {
        my $logdir = dirname($args->{"config_file"}) . "/../log";
        if (-d $logdir)
        {
            $args->{"default_appender_values"}{"path"} =
              abs_path("$logdir/$defaultLogName.log");
        }
        else
        {
            $args->{"default_appender_values"}{"path"} =
              "$Bin/$defaultLogName.log";
        }
        setConfigFile($args, $args->{"config_file"});
    }
    else
    {
        $args->{"default_appender_values"}{"path"} = "$Bin/$defaultLogName.log";
    }
    setAppenders($args, keys %appenders);
}

sub setConfigFile
{
    my $self        = shift;
    my $config_file = shift;

    $self->{"logger"}->logcroak("config_file path does not exist")
      unless defined($config_file)
          and -e $config_file;

    $config_file =~ s/\/$//;
    $self->{"config_file"} = abs_path($config_file);
    return 1;
}

sub _parse_config_file
{
    my $self          = shift;
    my $appenderNames = shift;

    $self->{"warnings"} = [];
    if (defined($self->{"config_file"}))
    {
        my $parser = Parse::RecDescent->new(
            qq(
		concept : predicate /\\[?/ constant /\\]?/ /\\s*/ attribute /=/ value
		{return \\%item}
		predicate : /APPENDERCFG|LOGGERCFG|/i
		constant : /[\\w\\\.-]+/i
		attribute : /PATH|LOGLEVEL|PERMISSIONS|MAXSIZE|ROLLOVERLIMIT|LAYOUT|/i
		value : /"/ pattern | variable
		variable : /(\\w|\\/|\\.|-)*/
		pattern : /[^"]*/
	    )
        );
        my $lineNumber = 0;

        open(FILE, $self->{"config_file"})
          or $self->{"logger"}->logcroak("config_file could not be opened.");
        foreach my $line (<FILE>)
        {
            $lineNumber++;
            next if $line =~ m/^\s*#/ or $line =~ m/^\s*\n?$/;
            my $info = $parser->concept($line)
              or chomp($line) and push(
                @{$self->{"warnings"}},
                "Skipping log4perl config line $lineNumber:$line"
              ) and next;
            my $predicate = lc($info->{predicate});
            my $constant  = lc($info->{constant});
            my $attribute = lc($info->{attribute});
            my $value     = $info->{value};

            if ($predicate eq "appendercfg"
                and exists($appenderNames->{$constant}))
            {

                if ($attribute eq "path" and $value =~ m/^\.\./)
                {
                    my $logdir  = dirname($self->{"config_file"}) . "/$value";
                    my $dirname = dirname($logdir);
                    unless (-d $dirname)
                    {
                        if ($dirname =~ /((\w|\/|\.|-)*)/)
                        {
                            $dirname = $1;
                        }
                        make_path($dirname)
                          or $self->{"logger"}
                          ->logcroak("Unable to make path for $dirname.\n");
                    }
                    $self->{"appenders"}{$constant}->{$attribute} =
                      abs_path($logdir);
                }
                elsif ($attribute eq "path")
                {
                    my $dirname = dirname($value);
                    unless (-d $dirname)
                    {
                        if ($dirname =~ /((\w|\/|\.|-)*)/)
                        {
                            $dirname = $1;
                        }
                        make_path($dirname)
                          or $self->{"logger"}
                          ->logcroak("Unable to make path for $dirname.\n");
                    }
                    $self->{"appenders"}{$constant}->{$attribute} = $value;
                }
                else
                {
                    $self->{"appenders"}{$constant}->{$attribute} = $value;
                }
            }
        }
        close(FILE);
    }
    else
    {
        push @{$self->{"warnings"}}, "No config_file set";

    }

    if (keys %{$self->{"appenders"}} == 0)
    {
        $self->{"appenders"}{"default"} = {};
    }

}

sub getConfigFile
{
    my $self = shift;
    return $self->{"config_file"};
}

sub setAppenders
{
    my ($self, @appenders) = @_;
    if (defined($self->{"appenders"}))
    {
        foreach my $appenderName (keys %{$self->{"appenders"}})
        {
            delete $self->{"appenders"}{$appenderName};
            if (exists $Log::Log4perl::Logger::APPENDER_BY_NAME{$appenderName})
            {
                $self->{"logger"}->remove_appender($appenderName);
            }
        }
    }
    my %lowerCaseAppenderNames = map { lc($_) => 1 } @appenders;
    _parse_config_file($self, \%lowerCaseAppenderNames);
    _appenders_init($self);
    _flush_config_warnings($self);
    return 1;
}

sub _flush_config_warnings
{
    my $self = shift;
    foreach my $warning (@{$self->{"warnings"}})
    {
        $self->{"logger"}->warn("$warning\n");
    }
}

sub _appenders_init
{
    my $self = shift;

    foreach my $appenderName (keys %{$self->{"appenders"}})
    {
        my $path = _AppenderAttribute($self, $appenderName, "path");
        if ($path =~ /((\w|\/|\.|-)*)/)
        {
            $path = $1;
        }
        my $permissions =
          _AppenderAttribute($self, $appenderName, "permissions");
        $permissions = oct($permissions);
        if ($permissions =~ /(\d+)/)
        {
            $permissions = $1;
        }
        my $layout = Log::Log4perl::Layout::PatternLayout->new(
            _AppenderAttribute($self, $appenderName, "layout"));

        my $appender = Log::Log4perl::Appender->new(
            "Log::Dispatch::FileRotate",
            name        => $appenderName,
            filename    => $path,
            permissions => $permissions,
            mode        => "append",
            size        => _AppenderAttribute($self, $appenderName, "maxsize"),
            min_level   => _AppenderAttribute($self, $appenderName, "loglevel"),
            max => _AppenderAttribute($self, $appenderName, "rolloverlimit")
        );

        $appender->layout($layout);
        $self->{"logger"}->add_appender($appender);
    }
}

sub _AppenderAttribute
{
    my $self     = shift;
    my $appender = shift;
    my $var      = shift;
    return undef unless defined $var and defined $appender;
    my $appenderSpecificVar = $self->{"appenders"}{$appender}->{$var};
    my $defaultVar          = $self->{"default_appender_values"}{$var};
    unless (defined($appenderSpecificVar) or defined($defaultVar))
    {
        $self->{"logger"}->logcroak(
            "$var not defined for the appender $appender. " . Dumper($self));
    }

    return defined($appenderSpecificVar) ? $appenderSpecificVar : $defaultVar;
}

sub getAppenders
{
    my $self = shift;
    return $self->{"appenders"};
}

sub setLogger
{
    my $self   = shift;
    my $logger = shift;
    if ($logger->isa("Log4perl"))
    {
        if (defined($self->{"logger"}))
        {
            Log::Log4perl->remove_logger($self->{"logger"});
        }
        $self->{"logger"} = $logger;
        return 1;
    }
    else
    {
        $self->{"logger"}->logcroak("Logger is not child type of Log4perl");
    }
}

sub getLogger
{
    my $self = shift;
    return $self->{"logger"};
}

sub debug
{
    my $self = shift;
    $self->{"logger"}->debug(@_);
}

sub info
{
    my $self = shift;
    $self->{"logger"}->info(@_);
}

sub warn
{
    my $self = shift;
    $self->{"logger"}->warn(@_);
}

sub error
{
    my $self = shift;
    $self->{"logger"}->error(@_);
}

sub fatal
{
    my $self = shift;
    $self->{"logger"}->fatal(@_);
}

sub is_debug
{
    my $self = shift;
    $self->{"logger"}->is_debug(@_);
}

sub is_info
{
    my $self = shift;
    $self->{"logger"}->is_info(@_);
}

sub is_warn
{
    my $self = shift;
    $self->{"logger"}->is_warn(@_);
}

sub is_error
{
    my $self = shift;
    $self->{"logger"}->is_error(@_);
}

sub is_fatal
{
    my $self = shift;
    $self->{"logger"}->is_fatal(@_);
}

sub level
{
    my $self = shift;
    $self->{"logger"}->level(@_);
}

sub more_logging
{
    my $self = shift;
    $self->{"logger"}->more_logging(@_);
}

sub less_logging
{
    my $self = shift;
    $self->{"logger"}->less_logging(@_);
}

sub inc_level
{
    my $self = shift;
    $self->{"logger"}->inc_level(@_);
}

sub dec_level
{
    my $self = shift;
    $self->{"logger"}->dec_level(@_);
}

sub logwarn
{
    my $self = shift;
    $self->{"logger"}->logwarn(@_);
}

sub logprint
{
    my $self = shift;
    $self->{"logger"}->info(@_);
    print join(" ", @_);
}

sub logcarp
{
    my $self = shift;
    $self->{"logger"}->logcarp(@_);
}

sub logcluck
{
    my $self = shift;
    $self->{"logger"}->logcluck(@_);
}

sub logcroak
{
    my $self = shift;
    $self->{"logger"}->logcroak(@_);
}

sub logconfess
{
    my $self = shift;
    $self->{"logger"}->logconfess(@_);
}

sub logdie
{
    my $self = shift;
    $self->{"logger"}->logdie(@_);
}

1;
