################################################################################
#
# Utility subroutines for Moab tools
#
#   Copyright (C) 2005-2014 by Adaptive Computing Enterprises, Inc. All Rights Reserved.
#
################################################################################
package Moab::Tools;

#use strict;
#use warnings;
use Exporter;
use Fcntl qw(:DEFAULT :flock);
use File::Basename;
use FindBin;

our $VERSION = '1.00_00';
    $VERSION = eval $VERSION;

our @ISA    = qw(Exporter);
our @EXPORT = qw
  (
  $homeDir
  &logPrint
  &logWarn
  &logDie
);
our @EXPORT_OK = qw
  (
  &readSpool
  &writeSpool
  &updateRow
  &deleteRow
  %monthToIndex
  %monthFromIndex
  %weekToIndex
  %weekFromIndex
);
our %EXPORT_TAGS = (
    spool   => [qw(&readSpool &writeSpool &updateRow &deleteRow)],
    date    => [qw(%monthToIndex %monthFromIndex %weekToIndex %weekFromIndex)],
    all     => [@EXPORT, @EXPORT_OK],
);

# Set the moab home directory
our $homeDir;
if ($ENV{MOABHOMEDIR})
{
    $homeDir = $ENV{MOABHOMEDIR};
}
if (! $homeDir && -r "/etc/moab.cfg")
{
    open GLOBALCONFIG, "< /etc/moab.cfg";
    $globalHomeDir = <GLOBALCONFIG>;
    $globalHomeDir =~ /MOABHOMEDIR\s+(\S+)/;
    $homeDir = $1;
    close GLOBALCONFIG;
}
if (! $homeDir && -x "${FindBin::Bin}/../bin/showq")
{
    open ABOUTMOAB, "${FindBin::Bin}/../bin/showq --about 2>&1 |";
    foreach my $line (<ABOUTMOAB>)
    {
        if ($line =~ /cfgdir=(\S+)/)
        {
            $homeDir = $1;
            last;
        }
    }
    close ABOUTMOAB;
}
unless ($homeDir)
{
    $homeDir = "/opt/moab";
}
$ENV{MOABHOMEDIR} = $homeDir;

# ----------------------------------------------------------------------------
# logPrint($message);
# ----------------------------------------------------------------------------

# Print message to the log file
sub logPrint
{
    my @message = @_;

    my $logName = basename($0);
    $logName =~ s/\.[^\.]+\.pl//;
    my $logFile = "${homeDir}/log/${logName}.log";

    # Roll the logs if necessary
    logRoll($logFile);

    # Print message to the log file
    if (open LOG, ">>$logFile")
    {
        my $time = scalar localtime(time);
        print LOG "\n$time ", @_;
    }
    close LOG;
}

# ----------------------------------------------------------------------------
# logWarn($message);
# ----------------------------------------------------------------------------

# Print message to the log file and to STDERR
sub logWarn
{
    my @message = @_;

    my $logName = basename($0);
    $logName =~ s/\.[^\.]+\.pl//;
    my $logFile = "${homeDir}/log/${logName}.log";

    # Roll the logs if necessary
    logRoll($logFile);

    #Print message to STDERR
    print STDERR @_;

    # Print message to the log file
    if (open LOG, ">>$logFile")
    {
        my $time = scalar localtime(time);
        print LOG "\n$time ", @_;
        close LOG;
    }
}

# ----------------------------------------------------------------------------
# logDie($message);
# ----------------------------------------------------------------------------

# Print message to the log file and then die
sub logDie
{
    my @message = @_;

    # Save off $? and $!
    my $rc = $?;
    my $ex = int $!;

    my $logName = basename($0);
    $logName =~ s/\.[^\.]+\.pl//;
    my $logFile = "${homeDir}/log/${logName}.log";

    # Roll the logs if necessary
    logRoll($logFile);

    # Print message to the log file
    if (open LOG, ">>$logFile")
    {
        my $time = scalar localtime(time);
        print LOG "\n$time ", @_;
        close LOG;
    }

    # Restore $? and $! and die
    $! = $ex;
    $? = $rc;
    die join ' ', @message;
}

# ----------------------------------------------------------------------------
# %data = readSpool($spoolFile);
# ----------------------------------------------------------------------------

# Read the spool, parse it into a hash of hashes, and return it
sub readSpool
{
    require 'shellwords.pl';
    my ($spoolFile) = @_;
    my %data = ();

    if (! -e "$spoolFile")
    {
        # Automatically create spool file to avoid error
        open SPOOL, ">>$spoolFile";
        close SPOOL;
    }

    if (open SPOOL, "<$spoolFile")
    {
        foreach my $line (<SPOOL>)
        {
            my @words;
            eval "
              no warnings;
              \@words = shellwords(\$line)
              ";
            next if $@;
            next unless scalar @words;
            my $rowId = shift @words;
            $data{$rowId} = ();
            foreach my $word (@words)
            {
                my ($key, $value) = split /=/, $word, 2;
                $data{$rowId}->{$key} = $value;
            }
        }
        close SPOOL;
    }
    else
    {
        logPrint("Unable to open spool file ($spoolFile) for reading: $!\n");
    }
    return %data;
}

# ----------------------------------------------------------------------------
# writeSpool($spoolFile, %data);
# ----------------------------------------------------------------------------

# Write out the current spool from the hash of hashes passed in
sub writeSpool
{
    my ($spoolFile, %data) = @_;
    if (open SPOOL, ">$spoolFile")
    {
        foreach my $rowId (sort keys %data)
        {
            print SPOOL $rowId;
            foreach my $key (sort keys %{$data{$rowId}})
            {
                my $value = $data{$rowId}->{$key};
                $value = "\"$value\"" if $value =~ / /;
                print SPOOL " $key=$value";
            }
            print SPOOL "\n";
        }
        close SPOOL;
    }
    else
    {
        logDie("Unable to open spool file ($spoolFile) for writing: $!\n");
    }
}

# ----------------------------------------------------------------------------
# updateRow($spoolFile, $rowId, %updates);
# ----------------------------------------------------------------------------

# Update a single data attribute in the spool file
# This routine uses file locking
sub updateRow
{
    my ($spoolFile, $rowId, %updates) = @_;
    require 'shellwords.pl';
    my %data = ();

    if (sysopen SPOOL, $spoolFile, O_RDWR | O_CREAT)
    {
        # Wait for a lock. First we do it nonblocking so we can log the wait
        unless (flock SPOOL, LOCK_EX | LOCK_NB)
        {
            logPrint("Spool file ($spoolFile) already locked; waiting...\n");
            flock SPOOL, LOCK_EX
              or logDie(
                "Unable to obtain exclusive lock on spool file ($spoolFile): $!\n"
              );
            logPrint("Lock obtained\n");
        }
        foreach my $line (<SPOOL>)
        {
            my @words;
            eval "
              no warnings;
              \@words = shellwords(\$line)
              ";
            next if $@;
            next unless scalar @words;
            my $rowId = shift @words;
            $data{$rowId} = ();
            foreach my $word (@words)
            {
                my ($key, $value) = split /=/, $word;
                $data{$rowId}->{$key} = $value;
            }
        }

        # Apply the updates
        foreach my $attribute (keys %updates)
        {
            $data{$rowId}->{$attribute} = $updates{$attribute};
        }

        # Truncate the file and write its modified contents back out
        truncate SPOOL, 0
          or logDie("Unable to truncate spool file ($spoolFile): $!\n");
        seek SPOOL, 0, 0;
        foreach my $rowId (sort keys %data)
        {
            my $line = $rowId;
            foreach my $key (sort keys %{$data{$rowId}})
            {
                my $value = $data{$rowId}->{$key};
                $value = "\"$value\"" if $value =~ / /;
                $line .= " $key=$value";
            }
            $line .= "\n";
            print SPOOL $line;
        }

        # Closing the file handle removes the file lock
        close SPOOL;
    }
    else
    {
        logDie("Unable to open spool file ($spoolFile) for update: $!\n");
    }
}

# ----------------------------------------------------------------------------
# deleteRow($spoolFile, $rowId);
# ----------------------------------------------------------------------------

# Delete a data entry in the spool file
# This routine uses file locking
sub deleteRow
{
    require 'shellwords.pl';
    my ($spoolFile, $rowId) = @_;
    if (open SPOOL, "+<$spoolFile")
    {
        # Wait for a lock. First we do it nonblocking so we can log the wait
        unless (flock SPOOL, LOCK_EX | LOCK_NB)
        {
            logPrint("Spool file ($spoolFile) already locked; waiting...\n");
            flock SPOOL, LOCK_EX
              or logDie(
                "Unable to obtain exclusive lock on spool file ($spoolFile): $!\n"
              );
            logPrint("Lock obtained\n");
        }
        foreach my $line (<SPOOL>)
        {
            my @words;
            eval "
              no warnings;
              \@words = shellwords(\$line)
              ";
            next if $@;
            next unless scalar @words;
            my $rowId = shift @words;
            $data{$rowId} = ();
            foreach my $word (@words)
            {
                my ($key, $value) = split /=/, $word;
                $data{$rowId}->{$key} = $value;
            }
        }

        # Delete the specified data row
        delete $data{$rowId};

        # Truncate the file and write its modified contents back out
        truncate SPOOL, 0
          or logDie("Unable to truncate spool file ($spoolFile): $!\n");
        seek SPOOL, 0, 0;
        foreach my $rowId (sort keys %data)
        {
            print SPOOL "$rowId";
            foreach my $key (sort keys %{$data{$rowId}})
            {
                my $value = $data{$rowId}->{$key};
                $value = "\"$value\"" if $value =~ / /;
                print SPOOL " $key=$value";
            }
            print SPOOL "\n";
        }

        # Closing the file handle removes the file lock
        close SPOOL;
    }
    else
    {
        logDie("Unable to open spool file ($spoolFile) for delete: $!\n");
    }
}

# ----------------------------------------------------------------------------
# logRoll($logFile);
# ----------------------------------------------------------------------------

# Roll the logs if necessary
sub logRoll
{
    my ($logFile) = @_;

    my $rollSize  = 10000000;
    my $rollCount = 7;

    if (-s $logFile && -s $logFile >= $rollSize)
    {
        for (my $i = $rollCount - 1; $i > 0; $i--)
        {
            my $fromFile = $logFile . '.' . $i;
            my $toFile = $logFile . '.' . ($i + 1);
            rename $fromFile, $toFile if -e $fromFile;
        }
        rename $logFile, $logFile . '.1';
        truncate $logFile, 0;
    }
}

# Map months of year to array indexes for timelocal
our %monthToIndex = (
    Jan => 0,
    Feb => 1,
    Mar => 2,
    Apr => 3,
    May => 4,
    Jun => 5,
    Jul => 6,
    Aug => 7,
    Sep => 8,
    Oct => 9,
    Nov => 10,
    Dec => 11
);

# Map array indexes to months of year for localtime
our %monthFromIndex = (
    0  => Jan,
    1  => Feb,
    2  => Mar,
    3  => Apr,
    4  => May,
    5  => Jun,
    6  => Jul,
    7  => Aug,
    8  => Sep,
    9  => Oct,
    10 => Nov,
    11 => Dec
);

# Map days of week to array indexes for timelocal
our %weekToIndex = (
    Sun => 0,
    Mon => 1,
    Tue => 2,
    Wed => 3,
    Thu => 4,
    Fri => 5,
    Sat => 6
);

# Map array indexes to days of week for localtime
our %weekFromIndex = (
    0 => Sun,
    1 => Mon,
    2 => Tue,
    3 => Wed,
    4 => Thu,
    5 => Fri,
    6 => Sat
);

1;
__END__
    
=head1 NAME
    
Moab::Tools - Perl extension for providing utility functions to Moab
    
=head1 SYNOPSIS

  use Moab::Tools;

=head1 DESCRIPTION
    
=head2 EXPORT
    
  $homeDir
  &logPrint
  &logWarn
  &logDie 

=head2 EXPORT_OK

  &readSpool
  &writeSpool
  &updateRow
  &deleteRow
  %monthToIndex
  %monthFromIndex
  %weekToIndex
  %weekFromIndex

=head1 SEE ALSO

Additional information can be found at http://www.adaptivecomputing.com/

=head1 AUTHOR

Adaptive Computing, Inc., E<lt>http://www.adaptivecomputing.com/E<gt>

=head1 COPYRIGHT AND LICENSE

Copyright (C) 2005-2014 by Adaptive Computing Enterprises, Inc. All Rights Reserved.

=cut
