################################################################################
#
#  Generic utility subroutines
#
#   Copyright (C) 2011-2014 by Adaptive Computing Enterprises, Inc. All Rights Reserved.
#
##################################################################################
package Moab::Utils::Utils;
use warnings;
use strict;
use FindBin;
use lib "/opt/moab/lib/perl5";
use Parse::RecDescent;
use Log::Log4perl qw(:easy);
use Cwd qw( abs_path );
use File::Basename;
use File::Path qw(make_path);
use Data::Dumper;

my %logAttributes = (
    "default" => {
        permissions   => 666,
        maxsize       => 10000000,
        rolloverlimit => 7,
        loglevel      => "info",
        layout        => "%d{yyyy-MM-dd	HH:mm:ss.SSS} %L  %-5p	%m%n"
    }
);
my %configAttributes = ();

Log::Log4perl->easy_init(
    {
        level  => $DEBUG,
        file   => ">>/dev/null",
        layout => $logAttributes{"default"}->{"layout"}
    }
);
my $log = get_logger();
my $type;
my $parsedConfig   = [];
my $preLogWarnings = [];

sub new
{
#this is an abstract parent class so it expects the blessed hash of the child class
    my $this = shift;
    my ($predicateRegex, $constantRegex, $attrRegex) = _makeConfigRegex(@_);
    $log->logcroak("type does not exist\n") unless defined($this->{type});
    $type = lc $this->{type};
    $log->logconfess("Utils not called with blessed hash")
      unless $this->isa(ref($this));
    $log->logcroak("cfg path does not exist\n")
      unless defined($this->{cfg})
          and -e $this->{cfg};
    $log->logcroak("Invalid type called to Utils\n")
      unless $type =~ m/^($constantRegex)$/i;
    $this->{cfg} =~ s/\/$//;
    $this->{cfg} = abs_path($this->{cfg});
    my $logdir = dirname($this->{cfg}) . "/../log";

    if (-d $logdir)
    {
        $logAttributes{"default"}->{"path"} =
          abs_path("$logdir/" . ref($this) . ".log");
    }
    else
    {
        $logAttributes{"default"}->{"path"} =
          dirname($this->{cfg}) . "/" . ref($this) . ".log";
    }
    _parseCfg($this->{cfg}, $predicateRegex, $constantRegex, $attrRegex);
    _logInit();
    _flushPreLogWarnings();
    return $this;
}

sub _flushPreLogWarnings
{
    foreach my $warn (@$preLogWarnings)
    {
        $log->info($warn);
    }
}

sub _makeConfigRegex
{
    my $predicates = shift;
    my $constants  = shift;
    my $attributes = shift;
    @$predicates = (@$predicates, ("LOG", "CONFIG"));
    @$constants = (@$constants, "default");
    @$attributes = (
        @$attributes,
        (
            "PATH",          "LOGLEVEL", "PERMISSIONS", "MAXSIZE",
            "ROLLOVERLIMIT", "LAYOUT"
        )
    );
    my $predicateRegex = join "|", @$predicates;
    my $constantRegex  = join "|", @$constants;
    my $attrRegex      = join "|", @$attributes;
    return ($predicateRegex, $constantRegex, $attrRegex);
}

sub _logInit
{
    my $path = _LogAttribute("path");
    if ($path =~ /((\w|\/|\.|-|:)*)/)
    {
        $path = $1;
    }
    my $permissions = _LogAttribute("permissions");
    $permissions = oct($permissions);
    if ($permissions =~ /(\d+)/)
    {
        $permissions = $1;
    }

    my $layout =
      Log::Log4perl::Layout::PatternLayout->new(_LogAttribute("layout"));

    my $appender = Log::Log4perl::Appender->new(
        "Log::Dispatch::FileRotate",
        name        => $type,
        filename    => $path,
        permissions => $permissions,
        mode        => "append",
        size        => _LogAttribute("maxsize"),
        min_level   => _LogAttribute("loglevel"),
        max         => _LogAttribute("rolloverlimit")
    );

    $appender->layout($layout);

    $log = get_logger($type);
    $log->add_appender($appender);
}

sub log
{
    return $log;
}

sub logprint
{
    my $class   = shift;
    my $message = shift;
    print "$message\n";
    $log->info($message);
}

sub getALLConfigAttributes
{
    return %configAttributes;
}

sub ConfigAttribute
{
    my $class = shift;
    my $var   = lc shift;
    return undef unless defined $var;
    my $typeSpecificVar = $configAttributes{$type}->{$var};
    my $defaultVar      = $configAttributes{default}->{$var};
    unless (defined($typeSpecificVar) or defined($defaultVar))
    {
        $log->logdie(
            "$var not defined for CONFIG[$type] in " . $class->{cfg} . "\n");
    }
    return defined($typeSpecificVar) ? $typeSpecificVar : $defaultVar;
}

sub _LogAttribute
{
    my $var = shift;
    return undef unless defined $var;
    my $typeSpecificVar = $logAttributes{$type}->{$var};
    my $defaultVar      = $logAttributes{default}->{$var};
    unless (defined($typeSpecificVar) or defined($defaultVar))
    {
        $log->logdie("$var not defined for LOG[$type] in config\n");
    }
    return defined($typeSpecificVar) ? $typeSpecificVar : $defaultVar;
}

sub _parseCfg
{
    my $cfgPath        = shift;
    my $predicateRegex = shift;
    my $constantRegex  = shift;
    my $attrRegex      = shift;

    my $parser = Parse::RecDescent->new(
        qq(
		concept : predicate /\\[?/ constant /\\]?/ /\\s*/ attribute /=/ value
		{return \\%item}
		predicate : /$predicateRegex/i
		constant : /$constantRegex/i
		attribute : /$attrRegex/i
		value : /"/ pattern | variable
		variable : /(\\w|\\/|\\.|-|:)*/
		pattern : /[^"]*/
	)
    );

    my $lineNumber = 0;
    my %hasStartJob;
    open(FILE, $cfgPath) or die "could not open $cfgPath\n";
    foreach my $line (<FILE>)
    {
        $lineNumber++;
        next if $line =~ m/^\s*#/ or $line =~ m/^\s*\n?$/;
        my $info = $parser->concept($line)
          or
          push(@$preLogWarnings, "skipping line $lineNumber of config:\n$line")
          and next;
        my $predicate = lc($info->{predicate});
        my $constant  = lc($info->{constant});
        my $attribute = lc($info->{attribute});
        my $value     = $info->{value};
        push @$parsedConfig, $info;

        if ($predicate eq "log")
        {
            if ($attribute eq "path" and $value =~ m/^\.\./)
            {
                my $logdir  = dirname($cfgPath) . "/$value";
                my $dirname = dirname($logdir);
                unless (-d $dirname)
                {
                    if ($dirname =~ /((\w|\/|\.|-)*)/)
                    {
                        $dirname = $1;
                    }
                    make_path($dirname)
                      or die
                      "Error: nami.cfg line $lineNumber unable to make path for log.\n";
                }
                $logAttributes{$constant}->{$attribute} = abs_path($logdir);
            }
            elsif ($attribute eq "path")
            {
                my $dirname = dirname($value);
                unless (-d $dirname)
                {
                    if ($dirname =~ /((\w|\/|\.|-)*)/)
                    {
                        $dirname = $1;
                    }
                    make_path($dirname)
                      or die
                      "Error: nami.cfg line $lineNumber unable to make path for log.\n";
                }
                $logAttributes{$constant}->{$attribute} = $value;
            }
            else
            {
                $logAttributes{$constant}->{$attribute} = $value;
            }
        }
        elsif ($predicate eq "config")
        {
            $configAttributes{$constant}->{$attribute} = $value;
        }
    }
    close(FILE);
}

sub _logParseInfo
{
    foreach my $info (@$parsedConfig)
    {
        my $line = "";
        foreach (keys %$info)
        {
            $line .= uc($_) . "=" . $info->{$_} . " " unless m/__/;
        }
        $log->debug("$line") unless $line eq "";
    }
}

sub _getParsedConfig
{
    _logParseInfo;
    my $CONFIG = {};
    $CONFIG->{"logAttributes"}    = \%logAttributes;
    $CONFIG->{"configAttributes"} = \%configAttributes;
    return $CONFIG;
}

1;
