#!/usr/bin/python

#THIS SOFTWARE SCRIPT IS PROVIDED BY ADAPTIVE COMPUTING ENTERPRISES, INC. "AS IS" 
#AND IS FOR REFERENCE USE ONLY.  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT 
#NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR 
#PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ADAPTIVE COMPUTING ENTERPRISES, INC. 
#BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
#DAMAGES ARISING FROM THE USE THEREOF.

#==============================================================================
# Torque detects KNL hardware and identifies each nodes current KNL state.
# This script generates a knl.nodes.txt file from pbsnodes XML output
# containing a single line for each node reporting opsys=CLE_<numa>_<mcdram>
# in the "status" element (list of comma separated name/value pairs)
#
# For example, if pbsnodes reports "opsys=CLE_a2a_flat" in the status element,
# for nid00001, this script will generate:
#
# nid00001 OSLIST=CLE_a2a_flat,CLE_a2a_equal,CLE_a2a_split,CLE_a2a_cache,CLE_snc2_flat,CLE_snc2_equal,CLE_snc2_split,CLE_snc2_cache,CLE_snc4_flat,CLE_snc4_equal,CLE_snc4_split,CLE_snc4_cache,CLE_hemi_flat,CLE_hemi_equal,CLE_hemi_split,CLE_hemi_cache,CLE_quad_flat,CLE_quad_equal,CLE_quad_split,CLE_quad_cache
#
# pseudocode:
# pbsnodes_xml = pbsnodes -x
# for each node in pbsnodes_xml
#    if status contains opsys=CLE_*
#        print "<node> OSLIST=CLE_a2a_flat,CLE_a2a_equal,CLE_a2a_split,CLE_a2a_cache,CLE_snc2_flat,CLE_snc2_equal,CLE_snc2_split,CLE_snc2_cache,CLE_snc4_flat,CLE_snc4_equal,CLE_snc4_split,CLE_snc4_cache,CLE_hemi_flat,CLE_hemi_equal,CLE_hemi_split,CLE_hemi_cache,CLE_quad_flat,CLE_quad_equal,CLE_quad_split,CLE_quad_cache"
#==============================================================================

from __future__ import print_function

import sys, argparse, subprocess
import xml.etree.ElementTree as ET

knl_configurations = [
    'CLE_a2a_flat',
    'CLE_a2a_equal',
    'CLE_a2a_split',
    'CLE_a2a_cache',
    'CLE_snc2_flat',
    'CLE_snc2_equal',
    'CLE_snc2_split',
    'CLE_snc2_cache',
    'CLE_snc4_flat',
    'CLE_snc4_equal',
    'CLE_snc4_split',
    'CLE_snc4_cache',
    'CLE_hemi_flat',
    'CLE_hemi_equal',
    'CLE_hemi_split',
    'CLE_hemi_cache',
    'CLE_quad_flat',
    'CLE_quad_equal',
    'CLE_quad_split',
    'CLE_quad_cache',
]

#==============================================================================
# Define CLI options
#==============================================================================
parser = argparse.ArgumentParser()

output_file_help="Output file"
parser.add_argument(
    "outfile",
    help=output_file_help
)

knl_feature_name_help="A feature name to be added to KNL nodes. Useful in heterogeneous hardware environments. (default: 'knl')"
parser.add_argument(
    "--knl-feature-name",
    dest='knl_feature_name',
    default='knl',
    help=knl_feature_name_help
)

non_knl_feature_name_help="A feature name to be added to non-KNL nodes. Useful in heterogeneous hardware environments. (default: 'non-knl')"
parser.add_argument(
    "--non-knl-feature-name",
    dest='non_knl_feature_name',
    default='non-knl',
    help=non_knl_feature_name_help
)

args = parser.parse_args()

#==============================================================================
# Run pbsnodes -x
#==============================================================================

pbsnodes_xml = subprocess.Popen("pbsnodes -x", shell=True,
    stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
# Block until complete
results = None
results = pbsnodes_xml.communicate()
if (pbsnodes_xml.returncode != 0):
    print("ERROR: ", "Failed to call pbsnodes -x.", file=sys.stderr)
    if results[0]: print("ERROR: ", "stdout: %s" % results[0], file=sys.stderr)
    if results[1]: print("ERROR: ", "stderr: %s" % results[1], file=sys.stderr)
    sys.exit(pbsnodes_xml.returncode)

output = open(args.outfile, 'w')
oslist = "OSLIST=%s" % ','.join(knl_configurations)

tree = ET.fromstring(results[0])
for node in tree.findall('./Node'):
    if node.find('status') != None:
        if 'opsys=CLE_' in node.find('status').text:
            print(node.find('name').text, "cray_compute", args.knl_feature_name, oslist, file=output)
        else:
            print(node.find('name').text, "cray_compute", args.non_knl_feature_name, file=output)

output.close()

sys.exit(0)
