#!/usr/bin/python

#THIS SOFTWARE SCRIPT IS PROVIDED BY ADAPTIVE COMPUTING ENTERPRISES, INC. "AS IS" 
#AND IS FOR REFERENCE USE ONLY.  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT 
#NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR 
#PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ADAPTIVE COMPUTING ENTERPRISES, INC. 
#BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
#DAMAGES ARISING FROM THE USE THEREOF.

# Modification for Cray XC systems
#==============================================================================
# THIS SCRIPT HAS BEEN MODIFIED TO SERVE AS A "REFERENCE" SCRIPT FOR USE WITH
# CRAY XC SYSTEMS.  IT MAY REQUIRE MODIFICATION TO WORK WITH A SPECIFIC CRAY.
#
# This is a Knights Landing (KNL) Cray XC management "reference" script that
# executes capmc commands to reconfigure and reinit/reboot KNL compute nodes
# within a Moab System Job prior to running a user job.
#
# Each compute node has an OS name feature assigned to it. The OS name
# identifies a node's current KNL configuration.
#
# OS=CLE_<numa_configuration>_<mcdram_configuration>
#
# As an optimization, Moab will schedule all required nodes, first reserving as
# many available nodes already in the required KNL configuration using the
# priority function (PRIORITYF) defined in moab.cfg for the PRIORITY Node
# Allocation Policy. Doing so potentially reduces job start times, because Moab
# will use nodes already configured with the requested KNL configuration. A list
# of reserved nodes that are not in the required configuration will be passed to
# this KNL Management script within a Moab System Job. This script will
# reconfigure and reinit/reboot each node using the capmc CLI.
#
# The Moab System job will complete successfully and Moab will run the user job
# once all nodes reserved for the job have been rebooted and Torque is reporting
# the requested OS name. The Moab System Job will fail if all nodes are not
# rebooted and reporting the requested OS name within a configurable
# (PROVDURATION) amount of time.
#
# Configuring Moab: See the README.INSTALL file
#
# Moab will make a call similar to the following when one or more nodes
# allocated to a job do not meet the OS requirement:
#
# $MOABHOMEDIR/tools/knl/knl.reconfigure.reboot.py nid00022,nid00021 --set OS=CLE_a2a_cache
# 
# The above script invocation would effectively reconfigure and reinit/reboot
# the list of nodes to use the a2a NUMA configuration and the cache MCDRAM
# configuration.
#
# Note that PROVDURATION=1500 would not allow the Moab System Job to run longer
# than 25 minutes.
#==============================================================================

from __future__ import print_function

import sys, argparse, time, subprocess, json
from config import Config

knl_configurations = [
    'OS=CLE_a2a_flat',
    'OS=CLE_a2a_equal',
    'OS=CLE_a2a_split',
    'OS=CLE_a2a_cache',
    'OS=CLE_snc2_flat',
    'OS=CLE_snc2_equal',
    'OS=CLE_snc2_split',
    'OS=CLE_snc2_cache',
    'OS=CLE_snc4_flat',
    'OS=CLE_snc4_equal',
    'OS=CLE_snc4_split',
    'OS=CLE_snc4_cache',
    'OS=CLE_hemi_flat',
    'OS=CLE_hemi_equal',
    'OS=CLE_hemi_split',
    'OS=CLE_hemi_cache',
    'OS=CLE_quad_flat',
    'OS=CLE_quad_equal',
    'OS=CLE_quad_split',
    'OS=CLE_quad_cache',
]

#==============================================================================
# Define CLI options
#==============================================================================
parser = argparse.ArgumentParser()

nodes_help="one or more nodes (comma separated) to reconfigure and reboot"
parser.add_argument(
    "nodes",
    help=nodes_help
)

set_help_string = """Specify the OS name that represents the required NUMA and
    MCDRAM configuration. Example: --set <ATTR>=<VALUE>, where <ATTR>=<VALUE> is
    one of the following: %s""" % ', '.join(knl_configurations)
parser.add_argument(
    '--set',
    help=set_help_string,
    metavar="KNL OS name",
    dest="os",
    choices=knl_configurations
)

args = parser.parse_args()

#==============================================================================
# Setup the environment for capmc
#==============================================================================

# Config() sets OS_CACERT, OS_KEY, OS_CERT, and OS_SERVICE_URL environment
# variables required by capmc. Setting these environment variables in a Config
# class allows one or more scripts like this one to include the same os.environ
# initialization code by simply instantiating an instance of the config class.
cfg = Config()

#==============================================================================
# Parse and sanitize the input argument(s)
#==============================================================================

# Tokenize <ATTR>=<VALUE> (i.e. OS=<knl_os_name>) into attr_eq_value. The
# "choices" kwarg to arparse's add_argument validates all values passed to
# --set. Therefore, we are guaranteed to have an <ATTR> = OS and a value that is
# in the knl_configurations list above.
attr_eq_value = args.os.split('=')

# Set knl_os_name to <VALUE>
knl_os_name = attr_eq_value[1]

# Decompose knl_os_name into numa_configuration and mcdram_configuration
# The knl_os_name is in the following format:
# CLE_<numa_configuration>_<mcdram_configuration>
knl_arguments = (knl_os_name.split('_'))
numa_configuration = knl_arguments[1]
mcdram_configuration = knl_arguments[2]

#DEBUG ONLY:
print("KNL OS Name: %s" % knl_os_name, file=sys.stdout)
print("NUMA configuration: %s" % numa_configuration, file=sys.stdout)
print("MCDRAM configuration: %s" % mcdram_configuration, file=sys.stdout)

# args.nodes may come as a comma separated list, like: "node1,node2,node3"
# Reduce the list to just numbers.
# TODO: Decide if your system requires "node" or "nid" to be removed.
#       Uncomment one or both of the following at your discretion.
#args.nodes=args.nodes.replace("node","")
#args.nodes=args.nodes.replace("nid","")

#==============================================================================
# Define the campc commands
#==============================================================================

# Format the capmc commands
# The knl_node_on_command will be formatted later
knl_set_mcdram_cfg_command = "%s set_mcdram_cfg --nids %s --mode %s" % (
    cfg.capmcCommand, args.nodes, mcdram_configuration)
knl_set_numa_cfg_command = "%s set_numa_cfg --nids %s --mode %s" % (
    cfg.capmcCommand, args.nodes, numa_configuration)
node_reboot_reason = "Moab KNL reconfiguration to NUMA: %s MCDRAM: %s" % (
    numa_configuration, mcdram_configuration)
knl_node_reinit_command = "%s node_reinit --nids %s --reason \"%s\"" % (
    cfg.capmcCommand, args.nodes, node_reboot_reason)

#==============================================================================
# Reconfigure MCDRAM
#==============================================================================

#DEBUG ONLY:
print("knl_set_mcdram_cfg_command = %s" % knl_set_mcdram_cfg_command, file=sys.stdout)
reconfigure_mcdram_process = subprocess.Popen(knl_set_mcdram_cfg_command,
    shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
# Block until complete
results = None
results = reconfigure_mcdram_process.communicate()
if (reconfigure_mcdram_process.returncode != 0):
    print("ERROR: ",
          "Failed to reconfigure MCDRAM. CAPMC command issued: %s rc: %i" % (
              knl_set_mcdram_cfg_command,
              reconfigure_mcdram_process.returncode),
          file=sys.stderr)
    if results[0]: print("ERROR: ", "stdout: %s" % results[0], file=sys.stderr)
    if results[1]: print("ERROR: ", "stderr: %s" % results[1], file=sys.stderr)
    sys.exit(reconfigure_mcdram_process.returncode)

## DEBUG ONLY:
#results = reconfigure_mcdram_process.stdout.readlines()
#print results

#==============================================================================
# Reconfigure NUMA
#==============================================================================

#DEBUG ONLY:
print("knl_set_numa_cfg_command = %s" % knl_set_numa_cfg_command, file=sys.stdout)

reconfigure_numa_process = subprocess.Popen(knl_set_numa_cfg_command,
    shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
# Block until complete
results = None
results = reconfigure_numa_process.communicate()
if (reconfigure_numa_process.returncode != 0):
    print("ERROR: ",
          "Failed to reconfigure NUMA. CAPMC command issued: %s rc: %i" % (
              knl_set_numa_cfg_command,
              reconfigure_numa_process.returncode),
          file=sys.stderr)
    if results[0]: print("ERROR: ", "stdout: %s" % results[0], file=sys.stderr)
    if results[1]: print("ERROR: ", "stderr: %s" % results[1], file=sys.stderr)
    sys.exit(reconfigure_numa_process.returncode)

#==============================================================================
# Reboot nodes
# NOTE: This is an asynchronous operation. Therefore, execute it once, check
# for errors, and return. Do not wait for the node to reinit.
#
# Moab is a fire-and-forget architecture. This script will simply return
# successful if there are no errors. There is no need to wait/poll until the
# node(s) is(are) in a ready state.
#
# Once this script returns successful, the Resource Manager (i.e. Torque) is
# responsible to report node state to Moab (including OS/KNL state).
#==============================================================================

#DEBUG ONLY:
print("knl_node_reinit_command = %s" % knl_node_reinit_command, file=sys.stdout)

node_reinit_process = subprocess.Popen(knl_node_reinit_command, shell=True,
    stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
# Block until complete
results = None
results = node_reinit_process.communicate()
if (node_reinit_process.returncode != 0):
    print("ERROR: ",
          "Failed to reinit/reboot nodes. CAPMC command issued: %s rc: %i" % (
              knl_node_reinit_command,
              node_reinit_process.returncode),
          file=sys.stderr)
    if results[0]: print("ERROR: ", "stdout: %s" % results[0], file=sys.stderr)
    if results[1]: print("ERROR: ", "stderr: %s" % results[1], file=sys.stderr)
    sys.exit(node_reinit_process.returncode)

sys.exit(0)
