#! /usr/bin/perl -w
################################################################################
#
# Usage: license.mon.flexLM.pl
#
# Script for interfacing with flexLM license manager
# To configure set @FLEXlmCmds
# To troubleshoot, set LOGLEVEL (0 - 7, 7 is most verbose)
#
# Added support to aggregate multiple license managers and multiple commands
# Set a maximum for available and configured licenses
#
#   Copyright (C) 2005-2016 by Adaptive Computing Enterprises, Inc. All Rights Reserved.
#
################################################################################
use strict;
use Getopt::Long;
use Time::Local;
use Time::localtime;

# Need to limit so not greater then Moab MAX_INT
use constant MAX_AVAIL => 1e6;
use constant MAX_CONFIG => 1e6;

my $LOGLEVEL = 0;
my $NODENAME = "GLOBAL";

# Parse Command Line Arguments
my ($help);
my $usage = "Usage: $0\n";
GetOptions('help|?' => \$help) or die $usage;
die $usage if $help;

#
# Insert lmstat (or lmutil lmstat) command here with -a and -c (license file)
# Multiple lfpaths are supported
# lmstat -a -c port@host1<:port@host>
# my @FLEXlmCmds = ("/full/path/lmutil lmstat -a -c 345@host1:456@host2");
# Multiple commands, enclosed in double quotes and seperated by commas
# are also supported
#
# It may be necessary to set the LM_LICENSE_FILE environment variable when multiple files across
# multiple servers are used
#
# $ENV{'LM_LICENSE_FILE'}='@blue;@green';

my @FLEXlmCmds = ("SETME");

my $FLEXlmCmd;
my $title     = "";
my $config    = 0;
my $avail     = 0;
my $sw        = "";
my $firstTime = 1;

my %swC = ();    # sw configured # keys are sw titles
my %swA = ();    # sw available  # keys are sw titles

# auto-detect flexLM version/format

my $NOT_SET    = 0;
my $OLD_FORMAT = 1;
my $NEW_FORMAT = 2;

my $format = $NOT_SET;

my $tm = localtime;
my $epochTime =
  timelocal($tm->sec, $tm->min, $tm->hour, $tm->mday, $tm->mon, $tm->year);

my $licenseCounter = 0;

for $FLEXlmCmd (@FLEXlmCmds)
  {
  if ($FLEXlmCmd =~ "SETME")
    {
    if ($LOGLEVEL >= 0)
      {
      print STDERR
        "ERROR:  script has not been configured, set \@FLEXlmCmds and retry\n";
      }

    exit(1);
    } # END if ($FLEXlmCmd =~ "SETME"...

  my ($cmd) = split /\s+/, $FLEXlmCmd;

  unless (-x $cmd)
    {
    if ($LOGLEVEL >= 0)
      {
      print STDERR
        "ERROR:  license query command not executable ($cmd)\n";
      }

    exit(1);
    }    #"$FLEXlmCmd not executable!"); }

  open(FLEXLM, "$FLEXlmCmd |")
    or die "Unable to execute license query command: $!\n";

  if ($LOGLEVEL >= 1)
    {
    print STDERR "INFO:   executing command $FLEXlmCmd\n";
    }

  while (<FLEXLM>)
    {
    chomp($_);

    if ($_ =~ m/status on \w+ (\d+)\/(\d+)\/(\d+) (\d+):(\d+)/)
      {
      # HEADER: load server timestamp
      # FORMAT: 6/13/2005 12:10

      if ($LOGLEVEL >= 1)
        {
        print STDERR "INFO:   license server header parsed\n";
        }

      $epochTime = timelocal(0, $5, $4, $2, $1 - 1, $3);
      } # END if ($_ =~ m/status on \w+ (\d+)\/(\d+)\/(\d+) (\d+):(\d+)/...

    if (/Error:/)
      {
      # ignore error messages

      if ($LOGLEVEL >= 1)
        {
        print STDERR "INFO:   ignoring error line $_\n";
        }

      next;
      } # END if (/Error:/)

    if (($format == $NOT_SET) && ($_ =~ m/Users of.+licenses?/))
      {
      if ($_ =~ m/licenses? available/)
        {
        $format = $OLD_FORMAT;

        if ($LOGLEVEL >= 2)
          {
          print STDERR "INFO:   old format line detected\n";
          }
        } # END if ($_ =~ m/licenses? available/...
      elsif ($_ =~ m/licenses? issued/)
        {
        $format = $NEW_FORMAT;

        if ($LOGLEVEL >= 2)
          {
          print STDERR "INFO:   new format line detected\n";
          }
        } # END elsif ($_ =~ m/licenses? issued/...
      else
        {
        if ($LOGLEVEL >= 1)
          {
          print STDERR "ERROR:  invalid flexLM format detected\n";
          }

        exit(1);    # unrecognized format
        } # END else [ if ($_ =~ m/licenses? available/...
      } # END if (($format == $NOT_SET...
    elsif ((($_ =~ m/licenses? available/) && ($format != $OLD_FORMAT))
      || (($_ =~ m/licenses? issued/) && ($format != $NEW_FORMAT)))
      {
      if ($LOGLEVEL >= 1)
        {
        print STDERR "ERROR:  inconsistent formats in command output\n";
        }

      exit(1);      # switching formats not allowed
      } # END elsif ((($_ =~ m/licenses? available/...

    if ($format == $OLD_FORMAT)
      {
      if ($_ =~ m/Users of ([^:]+):.*Total of (\d+) licenses? available/)
        {
        # Write out last stats

        $avail = $config - $licenseCounter;

        if (($avail >= 0) && ($title ne ""))
          {
          $swA{$title} = 0 if not defined $swA{$title};
          # Limit the max
          $swA{$title} = $swA{$title} + $avail < MAX_AVAIL ?
            $swA{$title} + $avail : MAX_AVAIL;
          }

        $licenseCounter = 0;
        $avail          = 0;

        # start new stats

        $title  = $1;
        $config = $2;

        if ($config < 0)
          {
          next;
          }

        $swC{$title} = 0 if not defined $swC{$title};
          # Limit the max
          $swC{$title} = $swC{$title} + $config < MAX_CONFIG ?
            $swC{$title} + $config : MAX_CONFIG;
        } # END if ($_ =~ ...
      elsif ($_ =~ m/, start/)
        {
        $licenseCounter++;
        }
      } # END if ($format == $OLD_FORMAT...
    elsif ($format == $NEW_FORMAT)
      {
      if ($_ =~ m/Users of ([^:]+):.*Total of (\d+) licenses?.*Total of (\d+)/)
        {
        $title  = $1;
        $config = $2;
        $avail  = $config - $3;

        if (($config < 0) || ($avail < 0))
          {
          next;
          }

       $swC{$title} = 0 if not defined $swC{$title};
        # Limit the max
        $swC{$title} = $swC{$title} + $config < MAX_CONFIG ?
          $swC{$title} + $config : MAX_CONFIG;
        $swA{$title} = 0 if not defined $swA{$title};
        # Limit the max
        $swA{$title} = $swA{$title} + $avail < MAX_AVAIL ?
          $swA{$title} + $avail : MAX_AVAIL;
        } # END if ($_ =~ ...
      } # END elsif ($format == $NEW_FORMAT...
    } # END while (<FLEXLM>)
  close FLEXLM;
  } # END for $FLEXlmCmd (@FLEXlmCmds...

if ($format == $OLD_FORMAT)
  {
  # generate last stats

  if ($LOGLEVEL >= 2)
    {
    print STDERR "INFO:   creating termination stats\n";
    }

  $avail = $config - $licenseCounter;

  if ($avail >= 0)
    {
    # Limit the max
    $swA{$title} = $avail < MAX_AVAIL ? $avail : MAX_AVAIL;
    }

  $licenseCounter = 0;
  $avail          = 0;
  } # END if ($format == $OLD_FORMAT...

if ($LOGLEVEL >= 2)
  {
  print STDERR "INFO:   displaying license status\n";
  }

print "$NODENAME UPDATETIME=$epochTime STATE=idle";

$firstTime = 1;

foreach $sw (sort keys %swA)
  {
  if ($firstTime)
    {
    $firstTime = 0;

    print " ARES=";
    } # END if ($firstTime)
  else
    {
    print ",";
    }

  print "$sw:$swA{$sw}";
  } # END foreach $sw (sort keys %swA...

$firstTime = 1;

foreach $sw (sort keys %swC)
  {
  if ($firstTime)
    {
    $firstTime = 0;

    print " CRES=";
    } # END if ($firstTime)
  else
    {
    print ",";
    }

  print "$sw:$swC{$sw}";
  } # END foreach $sw (sort keys %swC...

print "\n";

if ($LOGLEVEL >= 2)
  {
  print STDERR "INFO:   complete\n";
  }
1;
__END__

=head1 NAME

=head1 SYNOPSIS

=head1 DESCRIPTION

=head2 EXPORT

=head2 EXPORT_OK

=head1 SEE ALSO

Additional information can be found at http://www.adaptivecomputing.com/

=head1 AUTHOR

Adaptive Computing, Inc., E<lt>http://www.adaptivecomputing.com/E<gt>

=head1 COPYRIGHT AND LICENSE

Copyright (C) 2005-2016 by Adaptive Computing Enterprises, Inc. All Rights Reserved.

=cut
