#!/usr/bin/perl
#############################################################################
#
# Usage: echo "<Request>...</Request>" | account.query.mam.pl
#
# This script may be invoked by Moab's Native Accounting Manager Interface
# in order to query account information.
#
# The Request XML is provided on standard input. A non-zero exit code is
# considered a failure. Error messages should be written to standard error.
#
#   Copyright (C) 2015-2016 by Adaptive Computing Enterprises, Inc. All Rights Reserved.
#
#############################################################################

use strict;
use warnings;
use lib qw(/opt/mam/lib);
use Error qw(:try);
use XML::LibXML;
use MAM;

# Read standard input into a string
my $input = <STDIN>;
chomp($input);

if ($log->is_info())
{
    $log->info("Invoked $0 with standard input: $input");
}

my $request = new MAM::Request();
try
{
    $request = parseRequest($input);
}
catch MAM::Exception with
{
    &logAndWarn("Unable to parse request: " . $_[0]->{-text});
    exit int($_[0]->{-value} / 10);
};

# Customize the request as desired

# Obtain response from MAM
my $response = $request->getResponse();
my $status   = $response->getStatus();
my $code     = $response->getCode();
my $message  = $response->getMessage();
my @data     = $response->getData();

if ($status eq "Success" || $status eq "Warning")
{
    $log->info("$0 received successful response from MAM: $message");
    print STDOUT '<Data>', map($_->toString(), @data), "</Data>\n";
    exit 0;
}
else
{
    $log->warn("$0 received failure response from MAM (code $code): $message");
    print STDERR "$message\n";
    exit(int($code / 10));
}

# ----------------------------------------------------------------------------
# $request = parseRequest($input);
# ----------------------------------------------------------------------------

# Parse the input string into a request object
sub parseRequest
{
    my ($input) = @_;

    # Parse input into an XML object
    my $parser = new XML::LibXML();
    my $doc;
    try
    {
        $doc = $parser->parse_string($input);
    }
    catch Error with
    {
        throw MAM::Exception('302',
            "Malformed XML.\nThe message on standard in cannot be parsed into valid XML: ("
              . $input
              . ')');
    };

    # Fail if root element is not Request
    my $root = $doc->getDocumentElement();
    if ($root->nodeName() ne 'Request')
    {
        throw MAM::Exception('308',
                "Invalid message type.\nThe root element is "
              . $root->nodeName()
              . ' but should be Request.');
    }

    # Instantiate a new request
    my $request = new MAM::Request();

    # Extract action from the request
    my $action = $root->getAttribute('action');
    if ($action)
    {
        $request->setAction($action);
    }
    else
    {
        throw MAM::Exception('312', 'No action specified in the request.');
    }

    # Extract objects from the request
    my @objects = $root->getChildrenByTagName('Object');
    unless (@objects)
    {
        throw MAM::Exception('311', 'No object specified in the request.');
    }
    foreach my $inObject (@objects)
    {
        my $name      = $inObject->textContent();
        my $alias     = $inObject->getAttribute('alias');
        my $join      = $inObject->getAttribute('join');
        my $outObject = new MAM::Object(name => $name);
        $outObject->setAlias($alias) if $alias;
        $outObject->setJoin($join)   if $join;
        $request->addObject($outObject);
    }

    # Extract selections if present
    foreach my $inSelection ($root->getChildrenByTagName('Get'))
    {
        my $object       = $inSelection->getAttribute('object');
        my $name         = $inSelection->getAttribute('name');
        my $op           = $inSelection->getAttribute('op');
        my $alias        = $inSelection->getAttribute('alias');
        my $outSelection = new MAM::Selection(name => $name);
        $outSelection->setObject($object) if $object;
        $outSelection->setOperator($op)   if $op;
        $outSelection->setAlias($alias)   if $alias;
        $request->addSelection($outSelection);
    }

    # Extract assignments if present
    foreach my $inAssignment ($root->getChildrenByTagName('Set'))
    {
        my $name  = $inAssignment->getAttribute('name');
        my $op    = $inAssignment->getAttribute('op');
        my $value = $inAssignment->textContent();
        my $outAssignment =
          new MAM::Assignment(name => $name, value => $value);
        $outAssignment->setOperator($op)  if $op;
        $request->addAssignment($outAssignment);
    }

    # Extract conditions if present
    foreach my $inCondition ($root->getChildrenByTagName('Where'))
    {
        my $object       = $inCondition->getAttribute('object');
        my $name         = $inCondition->getAttribute('name');
        my $subject      = $inCondition->getAttribute('subject');
        my $op           = $inCondition->getAttribute('op');
        my $value        = $inCondition->textContent();
        my $conj         = $inCondition->getAttribute('conj');
        my $group        = $inCondition->getAttribute('group');
        my $inCondition = new MAM::Condition(name => $name, value => $value);
        $inCondition->setObject($object)    if $object;
        $inCondition->setSubject($subject)  if $subject;
        $inCondition->setOperator($op)      if $op;
        $inCondition->setConjunction($conj) if $conj;
        $inCondition->setGroup($group)      if $group;
        $request->addCondition($inCondition);
    }

    # Extract options if present
    foreach my $inOption ($root->getChildrenByTagName('Option'))
    {
        my $name  = $inOption->getAttribute('name');
        my $op    = $inOption->getAttribute('op');
        my $value = $inOption->textContent();
        my $outOption =
          new MAM::Option(name => $name, value => $value);
        $outOption->setOperator($op)  if $op;
        $request->setOption($outOption);
    }

    $log->debug("Extracted the request ($request).");

    return $request;
}
