#!/usr/bin/perl
#############################################################################
#
# Usage: echo "<Request>...</Request>" | lien.delete.mam.pl
#
# This script may be invoked by Moab's Native Accounting Manager Interface
# in order to clean up (e.g. remove stale liens) after a job failure.
#
# The Request XML is provided on standard input. A non-zero exit code is
# considered a failure. Error messages should be written to standard error.
#
#   Copyright (C) 2015-2016 by Adaptive Computing Enterprises, Inc. All Rights Reserved.
#
#############################################################################

use strict;
use warnings;
use lib qw(/opt/mam/lib);
use Error qw(:try);
use XML::LibXML;
use MAM;

# Read standard input into a string
my $input = <STDIN>;
chomp($input);

if ($log->is_info())
{
    $log->info("Invoked $0 with standard input: $input");
}

my $request = new MAM::Request();
try
{
    $request = parseRequest($input);
}
catch MAM::Exception with
{
    &logAndWarn("Unable to parse request: " . $_[0]->{-text});
    exit int($_[0]->{-value} / 10);
};

# Customize the request as desired

# Obtain response from MAM
my $response = $request->getResponse();
my $status   = $response->getStatus();
my $code     = $response->getCode();
my $message  = $response->getMessage();

if ($status eq "Success")
{
    $log->info("$0 received successful response from MAM: $message");
    print STDERR "$message\n" if $message;
    exit 0;
}
else
{
    $log->warn("$0 received "
          . lc($status)
          . " response from MAM (code $code): $message");
    print STDERR "$message\n" if $message;
    exit(int($code / 10));
}

# ----------------------------------------------------------------------------
# $request = parseRequest($input);
# ----------------------------------------------------------------------------

# Parse the input string into a request object
sub parseRequest
{
    my ($input) = @_;

    # Parse input into an XML object
    my $parser = new XML::LibXML();
    my $doc;
    try
    {
        $doc = $parser->parse_string($input);
    }
    catch Error with
    {
        throw MAM::Exception('302',
            "Malformed XML.\nThe message on standard in cannot be parsed into valid XML: ("
              . $input
              . ')');
    };

    # Fail if root element is not Request
    my $root = $doc->getDocumentElement();
    if ($root->nodeName() ne 'Request')
    {
        throw MAM::Exception('308',
                "Invalid message type.\nThe root element is "
              . $root->nodeName()
              . ' but should be Request.');
    }

    # Instantiate a new request
    my $request = new MAM::Request();

    # Extract action from the request
    my $action = $root->getAttribute('action');
    if ($action)
    {
        $request->setAction($action);
    }
    else
    {
        throw MAM::Exception('312', 'No action specified in the request.');
    }

    # Extract objects from the request
    my @objects = $root->getChildrenByTagName('Object');
    unless (@objects)
    {
        throw MAM::Exception('311', 'No object specified in the request.');
    }
    foreach my $object (@objects)
    {
        my $name = $object->textContent();
        $request->addObject($name);
    }

    # Extract conditions if present
    foreach my $condition ($root->getChildrenByTagName('Where'))
    {
        my $name  = $condition->getAttribute('name');
        my $value = $condition->textContent();
        $request->addCondition($name, $value);
    }

    $log->debug("Extracted the request ($request).");

    return $request;
}
