#! /usr/bin/perl -wT
################################################################################
#
# Retry a failed usage charge
#
# File   :  mam-charge-retry
#
################################################################################
#
#    This file includes parts which are copyrighted by Adaptive Computing
#    Enterprises, Inc., and parts which are copyrighted by Battelle Memorial
#    Institute.
#
#    The terms and conditions for the use and redistribution of these parts
#    are governed by the Moab Accounting Manager License and the BSD License
#    respectively. Refer to the LICENSE file for details.
#
#    Copyright (C) 2006 - 2015-2016 by Adaptive Computing Enterprises, Inc.
#    All Rights Reserved.
#
#    The Moab Accounting Manager License specifies the terms and conditions
#    for use and redistribution.
#
#    The Moab Accounting Manager License applies to the Moab Accounting
#    Manager software offered by Adaptive Computing Enterprises, Inc.
#    By installing or using this software, Licensee accepts a non-exclusive
#    license from Adaptive Computing Enterprises, Inc. and is bound to accept
#    acknowledgement of and abide by the notices and conditions of the
#    Moab Accounting Manager License.
#
#    Copyright (C) 2003 - 2005 Pacific Northwest National Laboratory,
#    Battelle Memorial Institute. All rights reserved.
#
#    The BSD License specifies the terms and conditions for use and
#    redistribution.
#
#    Redistribution and use in source and binary forms, with or without
#    modification, are permitted provided that the following conditions
#    are met:
#
#    - Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
#    - Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
#    - Neither the name of Battelle nor the names of its contributors
#    may be used to endorse or promote products derived from this software
#    without specific prior written permission.
#
#    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
#    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
#    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
#    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
#    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
#    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
#    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
#    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
#    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
#    ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
#    POSSIBILITY OF SUCH DAMAGE.
#
################################################################################

use strict;
use vars qw($log $verbose @ARGV %supplement $quiet $VERSION);
use lib qw(/opt/mam/lib /opt/mam/lib/perl5);
use Getopt::Long 2.24 qw(:config no_ignore_case);
use autouse 'Pod::Usage' => qw(pod2usage);
use Error qw(:try);
use JSON;
use MAM;

Main:
{
    $log->debug("Command line arguments: @ARGV");

    # Parse Command Line Arguments
    my ($account,         $charge,       $class,      $cpuTime,
        $deleteOnSuccess, $description,  $disk,       $dryRun,
        $duration,        $endTime,      %extensions, $fileName,
        $fund,            $group,        $help,       $hours,
        $incremental,     $instance,     $itemize,    $lien,
        $machine,         $man,          $memory,     $network,
        $nodes,           $organization, $partition,  $procs,
        $qos,             $quoteId,      @rates,      $site,
        $stage,           $startTime,    $type,       $usageRecordId,
        $user,            $version,
    );

    GetOptions(
        'a=s'               => \$account,
        'c=s'               => \$class,
        'C=f'               => \$cpuTime,
        'd=s'               => \$description,
        'delete-on-success' => \$deleteOnSuccess,
        'dry-run'           => \$dryRun,
        'D=s'               => \$disk,
        'e=s'               => \$endTime,
        'f=i'               => \$fund,
        'filename=s'        => \$fileName,
        'g=s'               => \$group,
        'hours|h'           => \$hours,
        'incremental'       => \$incremental,
        'itemize'           => \$itemize,
        'j=i'               => \$usageRecordId,
        'J=s'               => \$instance,
        'l=i'               => \$lien,
        'm=s'               => \$machine,
        'M=f'               => \$memory,
        'n=s'               => \$network,
        'N=i'               => \$nodes,
        'o=s'               => \$organization,
        'p=s'               => \$partition,
        'P=s'               => \$procs,
        'q=i'               => \$quoteId,
        'Q=s'               => \$qos,
        'R=s'               => \@rates,
        'stage=s'           => \$stage,
        's=s'               => \$startTime,
        't=i'               => \$duration,
        'T=s'               => \$type,
        'u=s'               => \$user,
        'extension|X=s'     => \%extensions,
        'z=f'               => \$charge,
        'debug'             => \&MAM::Client::enableDebug,
        'site=s'            => \$site,
        'help'              => \$help,
        'man'               => \$man,
        'quiet'             => \$quiet,
        'verbose|v'         => \$verbose,
        'usage-record'      => \&MAM::Client::parseSupplement,
        'where'             => \&MAM::Client::parseSupplement,
        'option'            => \&MAM::Client::parseSupplement,
        'version'           => \$version,
    ) or pod2usage(2);

    # Display usage if necessary
    pod2usage(2) if $help;
    if ($man)
    {
        if ($< == 0)    # Cannot invoke perldoc as root
        {
            my $id = eval { getpwnam("nobody") };
            $id = eval { getpwnam("nouser") } unless defined $id;
            $id = -2                          unless defined $id;
            $<  = $id;
        }
        $> = $<;                         # Disengage setuid
        $ENV{PATH} = "/bin:/usr/bin";    # Untaint PATH
        delete @ENV{'IFS', 'CDPATH', 'ENV', 'BASH_ENV'};
        if ($0 =~ /^([-\/\w\.]+)$/) { $0 = $1; }    # Untaint $0
        else { die "Illegal characters were found in \$0 ($0)\n"; }
        pod2usage(-exitstatus => 0, -verbose => 2);
    }

    # Display version if requested
    if ($version)
    {
        print "$PRODUCT version $VERSION\n";
        exit 0;
    }

    # Use sole remaining argument as file name if not specified by argument
    if (@ARGV == 1 && ! defined $fileName)
    {
        $fileName = $ARGV[0];
    }
    elsif (@ARGV > 0) { pod2usage(2); }

    # Must specify file name
    pod2usage(2) unless defined $fileName;

    # Check file name for existence and readability
    if (! -r $fileName)
    {
        pod2usage(
            -message =>
              "Invalid filename option value \"$fileName\" (File does not exist or is not readable)",
            -verbose => 0
        );
    }

    # Untaint file name
    if ($fileName =~ /^([-\/\w\.]+)$/) { $fileName = $1; }
    else { die "Illegal characters were found in file name ($fileName)\n"; }

    # Read file contents into a scalar
    open CHARGE_FILE, "<$fileName"
      or die "Unable to open $fileName for reading: $!\n";
    my $fileContents = do { local $/; <CHARGE_FILE> };
    my $json         = decode_json $fileContents;
    my $requestXML   = $json->{'request'};
    my $request      = new MAM::Request();
    try
    {
        $request = parseRequest($requestXML);
    }
    catch MAM::Exception with
    {
        warn("Unable to parse request: " . $_[0]->{-text});
        exit int($_[0]->{-value} / 10);
    };

    # Modify request
    $request->setSite($site) if defined $site;
    MAM::Client::buildSupplements($request);
    my @data        = $request->getData();
    my $usageRecord = $data[0];
    $usageRecord->setProperty("Account", $account) if defined $account;
    $usageRecord->setProperty("Class",   $class)   if defined $class;
    $usageRecord->setProperty("CPUTime", $cpuTime) if defined $cpuTime;
    $usageRecord->setProperty("Disk",    $disk)    if defined $disk;
    $request->setAction('Quote') if $dryRun;
    $usageRecord->setProperty("Group",    $group)    if defined $group;
    $usageRecord->setProperty("Instance", $instance) if defined $instance;
    $usageRecord->setProperty("Lien",     $lien)     if defined $lien;
    $usageRecord->setProperty("Machine",  $machine)  if defined $machine;
    $usageRecord->setProperty("Memory",   $memory)   if defined $memory;
    $usageRecord->setProperty("Network",  $network)  if defined $network;
    $usageRecord->setProperty("Nodes",    $nodes)    if defined $nodes;
    $usageRecord->setProperty("Organization", $organization)
      if defined $organization;
    $usageRecord->setProperty("Partition",  $partition) if defined $partition;
    $usageRecord->setProperty("Processors", $procs)     if defined $procs;
    $usageRecord->setProperty("QualityOfService", $qos)     if defined $qos;
    $usageRecord->setProperty("Quote",            $quoteId) if defined $quoteId;
    $usageRecord->setProperty("Stage",            $stage)   if defined $stage;
    $usageRecord->setProperty("Type",             $type)    if defined $type;
    $usageRecord->setProperty("Id",   $usageRecordId) if defined $usageRecordId;
    $usageRecord->setProperty("User", $user)          if defined $user;

    foreach my $key (keys %extensions)
    {
        pod2usage(
            -message => "Invalid extension property name ($key)",
            -verbose => 0
        ) if $key !~ /^[a-zA-Z][\w\-]*$/;
        $usageRecord->setProperty($key, $extensions{$key});
    }
    $request->setOption("Charge",      $charge)      if defined $charge;
    $request->setOption("Description", $description) if defined $description;
    $request->setOption("Duration",    $duration)    if defined $duration;
    $request->setOption("EndTime",     $endTime)     if defined $endTime;
    $request->setOption("Fund",        $fund)        if defined $fund;
    $request->setOption("Incremental", "True")       if $incremental;
    $request->setOption("Itemize",     "True")       if $itemize;
    $request->setOption("ShowHours",   "True")       if $hours;
    $request->setOption("StartTime",   $startTime)   if defined $startTime;

    foreach my $rateList (@rates)
    {
        foreach my $rate (split /,/, $rateList)
        {
            $request->setOption("Rate", $rate);
        }
    }
    $log->info("Built request: ", $request->toString());

    # Obtain Response
    my $response = $request->getResponse();

    # Print out the response
    &MAM::Client::displayResponse($response);

    if ($deleteOnSuccess && $response->getStatus() eq 'Success' && ! $dryRun)
    {
        unlink $fileName or warn "Warning: Unable to delete $fileName: $!\n";
    }

    # Exit with status code
    my $code = $response->getCode();
    $log->info("$0 (PID $$) Exiting with status code ($code)");
    exit $code / 10;
}

# ----------------------------------------------------------------------------
# $request = parseRequest($input);
# ----------------------------------------------------------------------------

# Parse the input string into a request object
sub parseRequest
{
    my ($input) = @_;

    # Parse input into an XML object
    my $parser = new XML::LibXML();
    my $doc;
    try
    {
        $doc = $parser->parse_string($input);
    }
    catch Error with
    {
        throw MAM::Exception('302',
            "Malformed XML.\nThe message on standard in cannot be parsed into valid XML: ("
              . $input
              . ')');
    };

    # Fail if root element is not Request
    my $root = $doc->getDocumentElement();
    if ($root->nodeName() ne 'Request')
    {
        throw MAM::Exception('308',
                "Invalid message type.\nThe root element is "
              . $root->nodeName()
              . ' but should be Request.');
    }

    # Instantiate a new request
    my $request = new MAM::Request();

    # Extract action from the request
    my $action = $root->getAttribute('action');
    if ($action)
    {
        $request->setAction($action);
    }
    else
    {
        throw MAM::Exception('312', 'No action specified in the request.');
    }

    # Extract objects from the request
    my @objects = $root->getChildrenByTagName('Object');
    unless (@objects)
    {
        throw MAM::Exception('311', 'No object specified in the request.');
    }
    foreach my $object (@objects)
    {
        my $name = $object->textContent();
        $request->addObject($name);
    }

    # Extract options if present
    foreach my $option ($root->getChildrenByTagName('Option'))
    {
        my $name  = $option->getAttribute('name');
        my $op    = $option->getAttribute('op');
        my $value = $option->textContent();
        $request->setOption($name, $value, $op);
    }

    # Extract data from the request
    my @dataNodes = $root->getChildrenByTagName('Data');
    if (@dataNodes)
    {
        $request->setDataElement($dataNodes[0]);
    }

    $log->debug("Extracted the request ($request).");

    return $request;
}

##############################################################################

__END__

=head1 NAME

mam-charge-retry - retry a failed usage charge via a failed charge retry file

=head1 SYNOPSIS

B<mam-charge-retry> {[B<--filename>] I<retry_filename>} [B<-J> I<instance_name>] [B<-j> I<usage_record_id>] [B<-q> I<quote_id>] [B<-l> I<lien_id>] [B<-T> I<usage_record_type>] [B<-u> I<user_name>] [B<-g> I<group_name>] [B<-a> I<account_name>] [B<-o> I<organization_name>] [B<-c> I<class_name>] [B<-Q> I<quality_of_service>] [B<-m> I<machine_name>] [B<-N> I<nodes>] [B<-P> I<processors>] [B<-C> I<cpu_time>] [B<-M> I<memory>] [B<-D> I<disk>] [B<--stage> I<lifecycle_stage>] [B<-X>, B<--extension> I<property>=I<value>]... [B<-t> I<charge_duration>] [B<-s> I<charge_start_time>] [B<-e> I<charge_end_time>] [B<-d> I<charge_description>] [B<-z> I<charge_amount>] [B<-f> I<fund_id>] [B<--incremental>] [B<-R> I<charge_rate_name>[{I<charge_rate_value>}]=I<charge_rate_amount>,...]... [B<--hours>] [B<--itemize>] [B<--delete-on-success>] [B<--dry-run>] [B<--debug>] [B<--site> I<site_name>] [B<--help>] [B<--man>] [B<--quiet>] [B<--verbose>] [B<--version>]

=head1 DESCRIPTION

B<mam-charge-retry> is used to retry a failed charge from a retry file. The specified command-line options will override the original values contained in the failed charge file.

=head1 OPTIONS

=over 4

=item B<-a> I<account_name>

account name

=item B<-c> I<class_name>

class of queue used

=item B<-C> I<cpu_time>

CPU time used

=item B<-d> I<charge_description>

explanatory message for the charge. The overall usage description can be passed via the extension property option (B<-X> Description=I<description>).

=item B<--delete-on-success>

delete the retry file after a successful charge

=item B<--dry-run>

issue the retry as a quote rather than a charge

=item B<-D> I<disk>

amount of disk used

=item B<-e> I<charge_end_time>

end time for the charge in the format YYYY-MM-DD[ hh:mm:ss]|-Infinity|Infinity|Now. Defaults to now if omitted. The overall usage end time can be passed via the extension property option (B<-X> EndTime=I<end_time>).

=item B<-f> I<fund_id>

fund_id to charge

=item [B<--filename>] I<retry_filename>

filename containing the original charge request to retry

=item B<-g> I<group_name>

group name

=item B<--incremental>

any associated liens will be debited instead of removed

=item B<-j> I<usage_record_id>

usage record id for the charge (if already created with mam-create-usagerecord, mam-quote, mam-reserve or a previous mam-charge). This is used to charge an existing usage record if the instance name (e.g. job id) is ambiguous or if a usage has already been debited and you want to charge an additional amount to the same usage record.

=item B<-J> I<instance_name>

instance name (e.g. job id) for the charge (if known). This can sometimes be non-unique (such as when a resource manager recycles job ids), and does not always unambiguously identify a usage record to charge. In such cases, look up and specify the usage record id for the charge.

=item B<-l> I<lien_id>

lien id (used to match up the right usage record id and remove the correct lien if ambiguous)

=item B<-m> I<machine_name>

name of the cluster

=item B<-M> I<memory>

amount of memory used

=item B<-N> I<nodes>

number of nodes used

=item B<-o> I<organization_name>

organization name

=item B<-P> I<processors>

number of processors used

=item B<-q> I<quote_id>

quote used to determine charge rates

=item B<-Q> I<quality_of_service>

quality of service used

=item B<-R> I<charge_rate_name>[{I<charge_rate_value>}]=I<charge_rate_amount>[,I<charge_rate_name>[{I<charge_rate_value>}]=I<charge_rate_amount>...]

uses the specified charge rates in the charge. The specified rates override the general rates or rates guaranteed through a quote. Multiple charge rates may be passed to the -R option in a comma-delimited list. Alternatively, multiple -R options may be specified.

=item B<-s> I<charge_start_time>

start time for the charge in the format YYYY-MM-DD[ hh:mm:ss]|-Infinity|Infinity|Now. Defaults to now - duration if omitted. The overall usage start time can be passed via the extension property option (B<-X> StartTime=I<start_time>).

=item B<--stage> I<lifecycle_stage>

latest stage in the object's accounting lifecycle (e.g. Create, Start, Continue, End)

=item B<-t> I<charge_duration>

wallclock duration for the charge in seconds. The total wallclock time for the lifetime usage can be passed via the extension property option (B<-X> Duration=I<duration_in_seconds>).

=item B<-T> I<usage_record_type>

specifies the usage record type (Job, Reservation, etc.)

=item B<-u> I<user_name>

user name

=item B<-X>, B<--extension> I<property>=I<value>

extension property. Any number of extra usage properties may be specified with the charge.

=item B<-z> I<charge_amount>

specifies the charge amount if calculated externally

=item B<--hours>

display time-based credits in hours. For cases where the currency is measured in resource-seconds (like processor-seconds), the currency is divided by 3600 to display resource-hours.

=item B<--itemize>
 
returns the composite charge information in the response data. This must be used in conjunction with the B<--verbose> flag to display the data.

=item B<--debug>

log debug info to screen

=item B<--site> I<site_name>

obtain response from specified site

=item B<--help>

brief help message

=item B<--man>

full documentation

=item B<--quiet>

suppress headers and success messages

=item B<--verbose>

display modified object details

=item B<--version>

display product version

=back

=head1 AUTHOR

Scott Jackson, scottmo@adaptivecomputing.com

=cut

