#!/usr/bin/perl
#############################################################################
#
# Usage: echo "<Request>...</Request>" | usage.quote.mam.pl
#
# This script may be invoked by Moab's Native Accounting Manager Interface
# in order to check for sufficient credits when a job or reservation is
# initially created.
#
# The Request XML is provided on standard input. A non-zero exit code is
# considered a failure. Error messages should be written to standard error.
#
#   Copyright (C) 2015-2016 by Adaptive Computing Enterprises, Inc. All Rights Reserved.
#
#############################################################################

use strict;
use warnings;
use lib qw(/opt/mam/lib);
use Error qw(:try);
use XML::LibXML;
use MAM;

# Read standard input into a string
my $input = <STDIN>;
chomp($input);

if ($log->is_info())
{
    $log->info("Invoked $0 with standard input: $input");
}

my $request = new MAM::Request();
try
{
    $request = parseRequest($input);
}
catch MAM::Exception with
{
    &logAndWarn("Unable to parse request: " . $_[0]->{-text});
    exit int($_[0]->{-value} / 10);
};

# Customize the request as desired
#my $datum = ($request->getData())[0];
#if ($datum->getPropertyValue('Type') eq 'Job')
#{
#    my $procs = $datum->getPropertyValue('Processors') || '1';
#    $datum->setProperty('Processors', $procs * 2);
#    my $quoteDuration = $request->getOptionValue('Duration') || 0;
#    $request->setOption('Charge', $procs * $quoteDuration);
#}

# Obtain response from MAM
my $response = $request->getResponse();
my $status   = $response->getStatus();
my $code     = $response->getCode();
my $message  = $response->getMessage();
my @data     = $response->getData();

if ($status eq "Success")
{
    $log->info("$0 received successful response from MAM: $message");
    print STDOUT join '', '<Data>', map({$_->toString()} @data), '</Data>', "\n"
      if @data;
    print STDERR "$message\n" if $message;
    exit 0;
}
else
{
    $log->warn("$0 received "
          . lc($status)
          . " response from MAM (code $code): $message");
    print STDOUT join '', '<Data>', map({$_->toString()} @data), '</Data>', "\n"
      if @data;
    print STDERR "$message\n" if $message;
    exit(int($code / 10));
}

# ----------------------------------------------------------------------------
# $request = parseRequest($input);
# ----------------------------------------------------------------------------

# Parse the input string into a request object
sub parseRequest
{
    my ($input) = @_;

    # Parse input into an XML object
    my $parser = new XML::LibXML();
    my $doc;
    try
    {
        $doc = $parser->parse_string($input);
    }
    catch Error with
    {
        throw MAM::Exception('302',
            "Malformed XML.\nThe message on standard in cannot be parsed into valid XML: ("
              . $input
              . ')');
    };

    # Fail if root element is not Request
    my $root = $doc->getDocumentElement();
    if ($root->nodeName() ne 'Request')
    {
        throw MAM::Exception('308',
                "Invalid message type.\nThe root element is "
              . $root->nodeName()
              . ' but should be Request.');
    }

    # Instantiate a new request
    my $request = new MAM::Request();

    # Extract action from the request
    my $action = $root->getAttribute('action');
    if ($action)
    {
        $request->setAction($action);
    }
    else
    {
        throw MAM::Exception('312', 'No action specified in the request.');
    }

    # Extract objects from the request
    my @objects = $root->getChildrenByTagName('Object');
    unless (@objects)
    {
        throw MAM::Exception('311', 'No object specified in the request.');
    }
    foreach my $object (@objects)
    {
        my $name = $object->textContent();
        $request->addObject($name);
    }

    # Extract options if present
    foreach my $option ($root->getChildrenByTagName('Option'))
    {
        my $name  = $option->getAttribute('name');
        my $op    = $option->getAttribute('op');
        my $value = $option->textContent();
        $request->setOption($name, $value, $op);
    }

    # Extract data from the request
    my @dataNodes = $root->getChildrenByTagName('Data');
    if (@dataNodes)
    {
        $request->setDataElement($dataNodes[0]);
    }

    $log->debug("Extracted the request ($request).");

    return $request;
}
