#!/usr/bin/python
# ====================================================================
# Each class should be written with the follwing in mind
#  init() - Initial call to setup the object
#  run()  - Ran only if init passes
#  offline_mode() - Gather logs and files.
#  diag() - Diagnostics
# ====================================================================

# ====================================================================
# Import from libraries and validate Python version
# ====================================================================
import ACgeneral as ACgeneral
import subprocess
import re
import os
import shutil


class ACslurm:
    """Common base class for all SLURM diag pieces"""
    def __init__(self, full_logging):
        self.home_dir = "/usr/local/etc/"
        self.full_log_mode = full_logging
        self.is_running = False
        self.pid = 0
        self.log_dir = ""
        self.commands_detected = False
        self.commands_active = False
        self.output_dir = "/tmp/"
        self.test_cmd = "scontrol ping"

    # ====================================================================
    #  Initialize - Initial call to setup the object
    # ====================================================================
    def init(self, output_dir):
        self.output_dir = output_dir
        self.check_path_for_slurm_cmds(ACgeneral.cmd_exists)
        if self.commands_detected:
            self.validate_slurmcmds_are_active(self.test_cmd)
        self.check_for_slurm_pid()

    # ====================================================================
    # Run - Ran only if init passes
    # ====================================================================
    def run(self):
        if not os.path.exists(self.output_dir+"slurm"):
            os.makedirs(self.output_dir+"slurm")

        self.gather_slurm_diags()
        self.gather_slurm_files()

    # ====================================================================
    # Offline mode
    # ====================================================================
    def offline_mode(self):
        if self.commands_detected:
            if not os.path.exists(self.output_dir+"slurm"):
                os.makedirs(self.output_dir+"slurm")
            self.gather_slurm_files()

    # ====================================================================
    # Diag - Diagnostics
    # ====================================================================
    def diag(self):
        diag_print = "\n================= SLURM DIAGS ================="
        diag_print += "\nSLURM home : " + self.home_dir
        diag_print += "\nSLRUM running : " + str(self.is_running)
        diag_print += "\nSLUMR PID : " + str(self.pid)
        diag_print += "\nSLUMR log dir : " + self.log_dir
        diag_print += "\nSLURM detected : " + str(self.commands_detected)
        diag_print += "\nSLURM commands active : " + str(self.commands_active)
        diag_print += "\nSLURM output dir : " + self.output_dir
        diag_print += "\nSLURM test cmd : " + self.test_cmd
        return diag_print

    # ====================================================================
    # Called from init() - Check for a SLURM pid 
    # ==================================================================== 
    def check_for_slurm_pid(self):
        return_code, cmd_output = ACgeneral.run_and_return_output("ps hf -opid -C slurmctld | awk '{ print $1; exit }'")
        if return_code == 0 and cmd_output == '':
            self.commands_detected = False
        elif return_code == 0:
            self.pid = cmd_output.rstrip('\n')

    # ====================================================================
    # Called from init() - Checks if commands are in the path
    # ====================================================================
    def check_path_for_slurm_cmds(self, func_c):
        # This checks sbin commands
        if func_c("sinfo"):
            self.commands_detected = True

    # ====================================================================
    # Called from init() - Check if moab commands work
    # ====================================================================  
    def validate_slurmcmds_are_active(self, cmd_to_run):
        p = subprocess.Popen(cmd_to_run, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
        output, error = p.communicate()

        if output:
            match = re.search(r'UP', output)
            if match:
                self.commands_detected = True
                self.commands_active = True
        elif error:
            match = re.search(r'command not found', error)
            if match:
                print "The SLURM commands are not in the path!"
                self.commands_detected = False
                self.commands_active = False
            match = re.search(r'RESTORE SLURMCTLD DAEMON TO SERVICE', error)
            if match:
                print "SLURM appears to be shutdown or unresponsive. "
                print "I will attempt to gather offline diagnostics..."
                self.commands_detected = True
                self.commands_active = False

    # ====================================================================
    # SLURM - Diagnostics
    # ====================================================================
    def gather_slurm_diags(self):
        os.chdir(self.output_dir)
        scontrol_show_jobs = "scontrol show jobs"
        sinfo_v = "sinfo"
        squeue_l_v = "squeue -l -v"
        scontrol_show_partition = "scontrol show partition"
        scontrol_show_config = "scontrol show config"
        scontrol_ping = "scontrol ping"
        try:
            ACgeneral.run_and_save(scontrol_show_jobs, "scontrol_show_jobs.txt", self.output_dir+"slurm/")
        except:
            print("Unable to run: " + scontrol_show_jobs)
        try:
            ACgeneral.run_and_save(sinfo_v, "sinfo_v.txt", self.output_dir+"slurm/")
        except:
            print("Unable to run: " + sinfo_v)
        try:
            ACgeneral.run_and_save(squeue_l_v, "squeue_l_v.txt", self.output_dir+"slurm/")
        except:
            print("Unable to run: " + squeue_l_v)
        try:
            ACgeneral.run_and_save(scontrol_show_partition, "scontrol_show_partition.txt", self.output_dir+"slurm/")
        except:
            print("Unable to run: " + scontrol_show_partition)
        try:
            ACgeneral.run_and_save(scontrol_show_config, "scontrol_show_config.txt", self.output_dir+"slurm/")
        except:
            print("Unable to run: " + scontrol_show_config)
        try:
            ACgeneral.run_and_save(scontrol_ping, "scontrol_ping.txt", self.output_dir+"slurm/")
        except:
            print("Unable to run: " + scontrol_ping)

        return 0

    # ====================================================================
    # SLURM - Files
    # ====================================================================
    def gather_slurm_files(self):
        os.chdir(self.output_dir)

        slurm_conf = "/usr/local/etc/slurm.conf"
        wiki_conf = "/usr/local/etc/wiki.conf"
        if os.path.isfile(slurm_conf):
            try:
                shutil.copy(slurm_conf, self.output_dir+"slurm/")
            except:
                print("Unable to copy: " + slurm_conf)
        if os.path.isfile(wiki_conf):
            try:
                shutil.copy(wiki_conf, self.output_dir+"slurm/")
            except:
                print("Unable to copy: " + wiki_conf)

        if os.path.isfile("/proc/"+str(self.pid)+"/fd/5"):
            try:
                shutil.copy("/proc/"+str(self.pid)+"/fd/5", self.output_dir+"slurm/slurmctld.log")
            except:
                print("Unable to copy your slurmctld logs.")
        return 0
