#!/usr/bin/python
# ====================================================================
# Each class should be written with the follwing in mind
#  init() - Initial call to setup the object
#  run()  - Ran only if init passes
#  offline_mode() - Gather logs and files.
#  diag() - Diagnostics
# ====================================================================

# ====================================================================
# Import from libraries and validate Python version
# ====================================================================
import re
import subprocess
import threading
import ACgeneral as ACgeneral
import os
import shutil
import datetime


class ACtorque:
    """Common base class for all TORQUE diag pieces"""
    def __init__(self, today_date_time, full_logging, core_dumps):
        self.cwd = ""
        self.home_dir = "/var/spool/torque/"
        self.core_dumps = core_dumps
        self.is_running = False
        self.pid = 0
        self.bin_path = "/usr/local/bin/"
        self.sbin_path = "/usr/local/sbin/"
        self.manually_detected = False
        self.commands_detected = False
        self.commands_active = False
        self.bin_commands_to_run = {}
        self.sbin_commands_to_run = {}
        self.log_dir = ""
        self.full_log_mode = full_logging
        self.port = ""
        self.test_cmd = "qstat -Q"
        self.today_date_time = today_date_time
        self.output_dir = "/tmp/"

    # ====================================================================
    #  Initialize - Initial call to setup the object
    # ====================================================================
    """
     Takes:
           func_c as cmd_exists in ACgeneral
    """
    def init(self, output_dir):
        self.output_dir = output_dir
        try:
            os.makedirs(self.output_dir+"torque_logs")
        except:
            print("I was unable to make the output direcotry: " + self.output_dir+"torque_logs")
            sys.exit(1)

        #Capture cwd to save.
        self.cwd = os.getcwd()
        self.check_path_for_torque_cmds(ACgeneral.cmd_exists)
        if self.commands_detected:
            self.validate_torquecmds_are_active()
        self.check_for_torque_pid()
        self.determin_torque_home()
        self.determin_cmd_path()

    # ====================================================================
    # Offline mode
    # ====================================================================
    def offline_mode(self):
        self.gather_torque_files()
        if self.core_dumps is not None:
            self.gather_torque_cores()

    # ====================================================================
    # Run - Ran only if init passes
    # ====================================================================
    def run(self):
        self.gather_torque_diagnostics()
        self.gather_torque_files()

        if self.core_dumps is not None:
            self.gather_torque_cores()

    # ====================================================================
    # Diag - Diagnostics
    # ====================================================================
    def diag(self):
        diag_print = "\n================= TORQUE DIAGS ================="
        diag_print += "\nTORQUE Home: " + self.home_dir
        diag_print += "\nTORQUE Running: " + str(self.is_running)
        diag_print += "\nTORQUE log dir: " + str(self.log_dir)
        diag_print += "\nTORQUE PID: " + str(self.pid)
        diag_print += "\nTORQUE Bin: " + self.bin_path
        diag_print += "\nTORQUE SBin: " + self.sbin_path
        diag_print += "\nTORQUE Manually detected: " + str(self.manually_detected)
        diag_print += "\nTORQUE Detected: " + str(self.commands_detected)
        diag_print += "\nTORQUE CMD active: " + str(self.commands_active)
        return diag_print

    # ====================================================================
    # Called from init() - Checks if commands are in the path
    # ====================================================================
    def check_path_for_torque_cmds(self, func_c):
        # This checks sbin commands
        if func_c("pbs_server"):
            # This checks for bin commands
            if func_c("qstat"):
                self.commands_detected = True

    # ====================================================================
    # Called from init() - Check if TORQUE commands work
    # ====================================================================  
    def validate_torquecmds_are_active(self):
        p = subprocess.Popen(self.test_cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
        output, error = p.communicate()

        if output:
            match = re.search(r'Queue', output)
            if match:
                self.commands_detected = True
                self.commands_active = True
        elif error:
            match = re.search(r'command not found', error)
            if match:
                print "The TORQUE commands are not in the path!"
                self.commands_detected = False
                self.commands_active = False
            match = re.search(r'cannot connect to server', error)
            if match:
                print "TORQUE appears to be shutdown or unresponsive. "
                print "I will attempt to gather offline diagnostics..."
                self.commands_detected = True
                self.commands_active = False

    # ====================================================================
    # Called from init() - Check for a TORQUE pid 
    # ==================================================================== 
    def check_for_torque_pid(self):
        # Check to see if TORQUE is running
        return_code, cmd_output = ACgeneral.run_and_return_output("ps hf -opid -C pbs_server | "
                                                                  "awk '{ print $1; exit }'")
        if return_code == 0 and cmd_output == '':
            print ("TORQUE process not found.")
        elif return_code == 0:
            self.pid = cmd_output.rstrip('\n')
            self.is_running = True

    # ====================================================================
    # Find TORQUEs home dir
    # ====================================================================
    def determin_torque_home(self):
        """
        :param current_directory:
        :param torque_object:
        :return:
        """
        if self.pid > 0:
            # use /proc/<pid>/cwd or  "pwd -P" "cd -P /proc/<pid>/cwd"
            path_to_cd = ""
            path_to_cd += str("/proc/")
            path_to_cd += str(self.pid)
            path_to_cd = path_to_cd.rstrip()
            # This will cd to server_priv
            # cwd -> /var/spool/torque/server_priv
            path_to_cd += str("/cwd/")
            os.chdir(path_to_cd)
            return_code, home_dir = ACgeneral.run_and_return_output("pwd -P")
            if return_code == 0:
                self.home_dir = str(home_dir)
                self.home_dir = self.home_dir.rstrip('\n')
                os.chdir(self.home_dir + "/../")
                self.home_dir = os.getcwd() + "/"
                os.chdir(self.output_dir)
                return 0
            else:
                print "Unable to determin TORQUE home directory. Assuming the default path."
                self.home_dir = None

            if self.home_dir is None:
                # Try known paths
                print "I do not see TORQUE running so I will try to determin the path manually."
                t_path_list = ["/var/spool/torque/", "/var/spool/pbs/", "/opt/torque/default/", "/opt/torque/"]
                for path_id in t_path_list:
                    if os.path.isdir(path_id):
                        self.home_dir = path_id
                        os.chdir(self.output_dir)
                        return 0
        return 0

    # ====================================================================
    # Called from init() - Check for TORQUE command path
    # ====================================================================
    def determin_cmd_path(self):
        if self.pid > 0 and self.commands_detected and self.commands_active:
            return_code, cmd_output = ACgeneral.run_and_return_output("which pbs_server")
            # Try parsing sbin dir out from which.
            if return_code == 0 and cmd_output == '':
                print ("I was not able to find sbin dir for TORQUE.")
            elif return_code == 0:
                self.sbin_path = re.sub('pbs_server', "", cmd_output)
                self.sbin_path = self.sbin_path.rstrip()
            # Try parsing bin dir out from which.
            return_code, cmd_output = ACgeneral.run_and_return_output("which qstat")
            if return_code == 0 and cmd_output == '':
                print ("I was not able to find bin dir for TORQUE client commands.")
            elif return_code == 0:
                self.bin_path = re.sub('qstat', "", cmd_output)
                self.bin_path = self.bin_path.rstrip()

        if self.pid > 0 and not self.commands_detected:
            print ("Looking for TORQUE client command path.")
            print ("Export your TORQUE binarie and client command path to skip auto detection.")
            proc_path = "/proc/" + str(self.pid)
            proc_path = proc_path.rstrip()
            proc_path += str("/")
            proc_path_to_test = "file -0 " + proc_path + "exe | awk '{print $5}'"
            return_code, bin_dir_tmp = ACgeneral.run_and_return_output(proc_path_to_test)
            if return_code == 0:
                sbin_dir = bin_dir_tmp
                sbin_dir = re.sub('`', "", sbin_dir)
                sbin_dir = re.sub("'", "", sbin_dir)
                sbin_dir = sbin_dir.strip()
                sbin_dir = sbin_dir[:-4]
                self.sbin_path = sbin_dir
                self.bin_path = re.sub("sbin", "bin", sbin_dir)
                self.bin_path = self.bin_path.strip()

        if self.pid == 0 and self.commands_detected and self.commands_active:
            print ("If this is a submit host please re-run on your HPC server.")
            print ("I will go ahead and gather client command output.")
            self.submit_host = True
        return 0

    # ====================================================================
    # Called from init() -> run() - Gather TORQUE files.
    # ====================================================================
    def gather_torque_files(self):
        """
        :return:
        """
        os.chdir(self.output_dir)
        trq_nodes_file = self.home_dir + "server_priv/nodes"
        trq_server_name = self.home_dir + "server_name"
        trq_torque_cfg = self.home_dir + "torque.cfg"
        trq_node_state = self.home_dir + "server_priv/node_status"
        trq_serverdb = self.home_dir + "server_priv/serverdb"
        trq_log_dir = self.home_dir + "server_logs/"
        self.log_dir = trq_log_dir

        # PID map for Moab
        torque_pid_map = "/proc/" + str(self.pid) + "/maps"
        torque_pid_limits = "/proc/" + str(self.pid) + "/limits"        

        if os.path.isfile(torque_pid_map):
            try:
                shutil.copy(torque_pid_map, self.output_dir+"/torque_pid_maps")
            except IOError, e:
                print "Unable to copy file. %s" % e
        if os.path.isfile(torque_pid_limits):
            try:
                shutil.copy(torque_pid_limits, self.output_dir+"/torque_pid_limits")
            except IOError, e:
                print "Unable to copy file. %s" % e

        if os.path.isfile(trq_nodes_file):
            shutil.copy(trq_nodes_file, self.output_dir)
        if os.path.isfile(trq_server_name):
            shutil.copy(trq_server_name, self.output_dir)
        if os.path.isfile(trq_torque_cfg):
            shutil.copy(trq_torque_cfg, self.output_dir)
        if os.path.isfile(trq_node_state):
            shutil.copy(trq_node_state, self.output_dir)
        if os.path.isfile(trq_serverdb):
            shutil.copy(trq_serverdb, self.output_dir)

        if self.full_log_mode and os.path.exists(trq_log_dir):
            ACgeneral.cp_file_pattern(trq_log_dir + "*", self.output_dir + "torque_logs/")
            return 0

        if os.path.exists(trq_log_dir):
            # Todays
            trq_logs_to_copy = trq_log_dir + self.today_date_time + "*"
            ACgeneral.cp_file_pattern(trq_logs_to_copy, self.output_dir + "torque_logs/")
            # Yesterdays
            trq_log_date_format_yesterday = str(int(self.today_date_time) - 1)
            trq_logs_to_copy = trq_log_dir + trq_log_date_format_yesterday + "*"
            ACgeneral.cp_file_pattern(trq_logs_to_copy, self.output_dir + "torque_logs/")
            # Daybefore Yesterday
            trq_log_date_format_day_before_yesterday = str(int(self.today_date_time) - 2)
            trq_logs_to_copy = trq_log_dir + trq_log_date_format_day_before_yesterday + "*"
            ACgeneral.cp_file_pattern(trq_logs_to_copy, self.output_dir + "torque_logs/")

        return 0

    # ====================================================================
    # Gather TORQUE command outputs and configs
    # ====================================================================
    def gather_torque_diagnostics(self):
        os.chdir(self.output_dir)
        commands_to_run = {"pbsnodes -a": "pbsnodes_-a.txt",
                           "qstat": "qstat.txt",
                           "qstat -f": "qstat_-f.txt",
                           "qstat -q": "qstat_-q.txt",
                           "qmgr -c 'p s'": "qmgr_-c_p_s.txt",
                           "qstat --about 2>&1": "qstat_--about.txt",
                           "pbs_server --about": "pbs_server_--about.txt",
                           "qstat -Q": "qstat_-Q.txt",
                           "qstat -a": "qstat_-a.txt",
                           "qstat -t": "qstat_-t.txt",
                           "qstat -s": "qstat_-s.txt"}
        ldd_cmd = "none"
        if ACgeneral.cmd_exists("pbs_server"):
            ldd_cmd = "ldd `which pbs_server`"
        elif os.path.isfile(self.sbin_path + "pbs_server"):
            ldd_cmd = "ldd " + self.sbin_path + "pbs_server"
        if ldd_cmd != "none":
            ACgeneral.run_and_save(ldd_cmd, "ldd_pbs_server.txt", self.output_dir)
        # Thread off TORQUE commands
        if self.manually_detected:
            for cmd, out_file in commands_to_run.iteritems():
                match = re.search(r'pbs_server', cmd)
                if match:
                    t_thread = threading.Thread(target=ACgeneral.run_and_save,
                                                args=(self.sbin_path+cmd, out_file, self.output_dir), name="torque")
                    t_thread.name = "torque"
                    t_thread.daemon = True
                    t_thread.start()
                else:
                    t_thread = threading.Thread(target=ACgeneral.run_and_save,
                                                args=(self.bin_path+cmd, out_file, self.output_dir), name="torque")
                    t_thread.name = "torque"
                    t_thread.daemon = True
                    t_thread.start()
        else:
            for cmd, out_file in commands_to_run.iteritems():
                t_thread = threading.Thread(target=ACgeneral.run_and_save,
                                            args=(cmd, out_file, self.output_dir), name="torque")
                t_thread.name = "torque"
                t_thread.daemon = True
                t_thread.start()

        if os.path.exists(self.home_dir):
            # Gather ls -hR for directory listing and size
            ACgeneral.run_and_save("ls -lhR " + self.home_dir, "torque_ls_-lhR.txt", self.output_dir)

        proc_path = "/proc/" + str(self.pid)
        proc_path = proc_path.rstrip()
        proc_path += str("/")
        if os.path.exists(proc_path):
            # Gather ls -hR for directory listing and size
            ACgeneral.run_and_save("ls -lhR " + proc_path, "torque_proc_ls_-lhR.txt", self.output_dir)


        return 0

    # ====================================================================
    # Check for core files.
    # ====================================================================
    def gather_torque_cores(self):
        list_cores = ACgeneral.find_file_from_pattern("*core*", self.home_dir)
        # Try to make the torque_core folder in outpur dir
        try:
            os.makedirs(self.output_dir+"torque_core")
        except:
            print("I was unable to make the output direcotry: " + self.output_dir + "torque_core")
            return 1

        if list_cores and self.core_dumps == "A":
            for cor in list_cores:
                if os.path.isfile(cor):
                    try:
                        shutil.copy(cor, self.output_dir+"torque_core/")
                    except:
                        e = sys.exc_info()[0]
                        print("There was a failure with trying to copy: %s" % e)
        elif list_cores and ACgeneral.str_is_int(self.core_dumps):
            if int(self.core_dumps) == 0:
                return 0
            files_and_dates_base = ACgeneral.return_file_and_mtime(self.home_dir, "*core*")
            if files_and_dates_base:
                self.copy_torque_cores(files_and_dates_base)

    # ====================================================================
    # Copy core files. Called from gather_moab_cores
    # ====================================================================
    def copy_torque_cores(self, files_and_dates):
        tmp_core_dump_count = int(self.core_dumps)
        core_count = 1
        for file_name, file_mdate in files_and_dates.iteritems():
            # Date format
            torque_core_date_format = str(int(self.today_date_time) - tmp_core_dump_count)
            if file_mdate >= torque_core_date_format:
                print "Copying core file: " + file_name
                if os.path.isfile(file_name):
                    # Copy core
                    try:
                        shutil.copy(file_name, self.output_dir+"torque_core/")
                    except:
                        e = sys.exc_info()[0]
                        print ("I was unable to copy " + file_name + " %s" % e)
                    # Grab the backtrace
                    try:
                        ACgeneral.run_and_save("gdb -q -n -ex bt -batch " + self.sbin_path+"pbs_server " +
                                               file_name + " 2>&1 | grep -v 'No stack.'\$",
                                               str(core_count)+"_bt.txt", self.output_dir+"torque_core/")
                        ACgeneral.run_and_save("gdb -q -n -ex 'bt full' -batch "+self.sbin_path+"pbs_server " +
                                               file_name + " 2>&1 | grep -v 'No stack.'\$",
                                               str(core_count)+"_bt_full.txt", self.output_dir+"torque_core/")
                    except:
                        e = sys.exc_info()[0]
                        print("I was unable to gather a backtrace from: " + file_name + " %s" % e)
