#! /usr/bin/perl -w
################################################################################
#
# showevents - list historical moab events
#
#   Copyright (C) 2006-2016 by Adaptive Computing Enterprises, Inc. All Rights Reserved.
#
################################################################################

use strict;
use FindBin;
use lib "/opt/moab/lib/perl5", "/opt/moab/etc";
use lib "${FindBin::Bin}/../../lib/perl5", "${FindBin::Bin}/../../etc";
use Getopt::Long 2.24 qw(:config no_ignore_case);
use Time::Local;
use Moab::Tools qw(:DEFAULT :date);    # Must be before required config file
use autouse 'Pod::Usage' => qw(pod2usage);
BEGIN { require "config.moab.pl"; }
our ($homeDir, $logLevel);

my %fieldIndex = (
    "Discard"           => 0,
    "EventTime"         => 1,
    "Time"              => 1,
    "TimeStamp"         => 1,
    "Object"            => 2,
    "ObjectType"        => 2,
    "ObjectId"          => 3,
    "JobId"             => 3,
    "Id"                => 3,
    "Event"             => 4,
    "Nodes"             => 5,
    "Tasks"             => 6,
    "Procs"             => 6,
    "User"              => 7,
    "Group"             => 8,
    "WallclockLimit"    => 9,
    "CompletionState"   => 10,
    "Status"            => 10,
    "Class"             => 11,
    "Queue"             => 11,
    "SubmitTime"        => 12,
    "DispatchTime"      => 13,
    "StartTime"         => 14,
    "CompletionTime"    => 15,
    "EndTime"           => 15,
    "Arch"              => 16,
    "Architecture"      => 16,
    "Opsys"             => 17,
    "OperatingSystem"   => 17,
    "MemCmp"            => 18,
    "Memory"            => 19,
    "DiskCmp"           => 20,
    "Disk"              => 21,
    "Features"          => 22,
    "SystemQueueTime"   => 23,
    "AllocatedTasks"    => 24,
    "TPN"               => 25,
    "ProcsPerNode"      => 25,
    "TasksPerNode"      => 25,
    "QOS"               => 26,
    "Flags"             => 27,
    "Account"           => 28,
    "Executable"        => 29,
    "Extra"             => 30,
    "RMExtension"       => 30,
    "Bypass"            => 31,
    "PSUtil"            => 32,
    "Partition"         => 33,
    "ProcsPerTask"      => 34,
    "MemPerTask"        => 35,
    "DiskPerTask"       => 36,
    "SwapPerTask"       => 37,
    "StartLimitTime"    => 38,
    "EndLimitTime"      => 39,
    "DeadlineTime"      => 39,
    "AllocatedHostList" => 40,
    "HostList"          => 40,
    "RM"                => 41,
    "RequiredHostList"  => 42,
    "Reservation"       => 43,
    "SimData"           => 44,
    "Constraints"       => 45,
    "Message"           => 46,
    "Cost"              => 47,
    "History"           => 48,
    "Utilization"       => 49,
    "Reserved1"         => 50,
    "Reserved2"         => 51,
    "Reserved3"         => 52,
    "Details"           => 53,
);

Main:
{
    logPrint("Command line arguments: @ARGV\n")
      if $logLevel;

    my %wiki_tbl;
    my @wiki_term;
    # network is not supported                  
    my $space = '';
    my $features;
    my (
              $executable,
              $extra,
              $extensions,
              $bypassCount,
              $PSUtil,
              $partitionName,
              # procsPerTask is not in WIKI
              $memoryPerTask,
              $diskPerTask,
              $swapPerTask,
              $activationTime,
              $deadlineTime,
              $allocatedHostlist,
              $resourceManager,
              $requestedHostList,
              $reservationId,
              $simData,
              $nodeConstraints,
              $messages,
              $cost,
              $history,
              $utilization,
              $estimateData,
              $complCodeIWDOut,
              $memoryStats,
              $cpuStats,
       );
    
    
    # Parse Command Line Arguments
    my (
        $account, $class, $group,  $help, $id,   $job,
        $man,     $num,   $object, $qos,  $show, $user
    );
    GetOptions(
        'account=s' => \$account,
        'class=s'   => \$class,
        'group=s'   => \$group,
        'help|?'    => \$help,
        'id=s'      => \$id,
        'j'         => \$job,
        'man'       => \$man,
        'num=s'     => \$num,
        'object=s'  => \$object,
        'qos=s'     => \$qos,
        'show=s'    => \$show,
        'user=s'    => \$user,
    ) or pod2usage(2);

    # Display usage if necessary
    pod2usage(0) if $help;
    if ($man)
    {
        if ($< == 0)    # Cannot invoke perldoc as root
        {
            my $id = eval { getpwnam("nobody") };
            $id = eval { getpwnam("nouser") } unless defined $id;
            $id = -2                          unless defined $id;
            $<  = $id;
        }
        $> = $<;                         # Disengage setuid
        $ENV{PATH} = "/bin:/usr/bin";    # Untaint PATH
        delete @ENV{'IFS', 'CDPATH', 'ENV', 'BASH_ENV'};
        if ($0 =~ /^([-\/\w\.]+)$/) { $0 = $1; }    # Untaint $0
        else { die "Illegal characters were found in \$0 ($0)\n"; }
        pod2usage(-exitstatus => 0, -verbose => 2);
    }

    # Use remaining argument as object id
    if (@ARGV)
    {
        if (@ARGV == 1 && !$id)
        {
            ($id) = @ARGV;
        }
        else
        {
            pod2usage(2);
        }
    }

    # Validate Fields
    my %fieldSize = ();
    if ($job)
    {
        $show = "TimeStamp,JobId,Event,User,Class,Account,Procs,Status"
          unless $show;
    }
    else
    {
        $show = "TimeStamp,Object,Id,Event,Details" unless $show;
    }
    my @show = split /,/, $show;
    foreach my $field (@show)
    {
        logDie("Invalid field ($field)\n") unless defined $fieldIndex{$field};
        $fieldSize{$field} = length $field;
    }

    # Build a sorted list of stats files
    my @eventsFiles = glob("${homeDir}/stats/events.*_*_*_*");
    @eventsFiles = sort by_date @eventsFiles;
    if ($num)
    {
        my $firstIndex = ($num <= @eventsFiles) ? @eventsFiles - $num : 0;
        my $lastIndex = $#eventsFiles;
        @eventsFiles = @eventsFiles[$firstIndex .. $lastIndex];
    }

    # Parse the stats files and populate @events
    my @events = ();
    foreach my $eventsFile (@eventsFiles)
    {
        open EVENTS, "< $eventsFile"
          or logDie(
            "Unable to open events file ($eventsFile) for reading: $!\ n
      "
          );
        foreach my $line (<EVENTS>)
        {
            chomp $line;
            my @fields = split /\s+/, $line;

            # Timestamp changed in Moab 5.4
            $fields[1] =~ s/:.*// if defined $fields[1];

            my (
                $hRTime,         $timestamp,       $objectType,
                $objectId,       $objectEvent,     $nodes,
                $tasks,          $userName,        $groupName,
                $wallclockLimit, $completionState, $className,
                $submissionTime, $dispatchTime,    $startTime,
                $completionTime, $network,         $architecture,
                $opsys,          $memcmp,          $memory,
                $diskcmp,        $disk,            $systemQueueTime,
                $allocatedTasks, $TPN,             $qosName,
                $flags,          $accountName,     @rest
            ) = @fields;
            next unless @fields >= 6;

            # TODO: Ensure this
            # force non-job records to not be WIKI
            $tasks = 0 if not defined $tasks;

            # WIKI format 
            if ($tasks =~ /=/)
            {
              %wiki_tbl = ();
              
              # WIKI terms start after the header
              for my $val (@fields[6..$#fields])
              {
                if ($val !~ /=/)
                {
                  print STDERR "WIKI term \'$val\' does not contain =\n";
                  next;
                }
                
                # limit for memory compare terms
                @wiki_term = split /=/, $val, 2;
                
                $wiki_tbl{$wiki_term[0]} =  $wiki_term[1];
                
              }
              
              $wiki_tbl{'OBJID'} = $objectId;
              
              $tasks = $wiki_tbl{'TASKS'};
              $userName = $wiki_tbl{'UNAME'};
              $groupName = $wiki_tbl{'GNAME'};
              $wallclockLimit = $wiki_tbl{'WCLIMIT'};
              $completionState = $wiki_tbl{'STATE'};
              $className = $wiki_tbl{'RCLASS'}; 
              $submissionTime = $wiki_tbl{'SUBMITTIME'};
              $dispatchTime = $wiki_tbl{'DISPATCHTIME'};
              $startTime = $wiki_tbl{'STARTTIME'};
              $completionTime = $wiki_tbl{'COMPLETETIME'};
              $architecture = $wiki_tbl{'ARCH'}; #
              $opsys = $wiki_tbl{'ROPSYS'};
              $memcmp = $wiki_tbl{'RMEMCMP'}; 
              $memory = $wiki_tbl{'DMEM'}; 
              $diskcmp = $wiki_tbl{'RDISKCMP'}; 
              $disk = $wiki_tbl{'RDISK'};  #
              # features are not in the field variables
              $systemQueueTime = $wiki_tbl{'SYSTEMQUEUETIME'};
              $allocatedTasks = $wiki_tbl{'TASKS'};
              $TPN = $wiki_tbl{'TPN'}; #
              $qosName = $wiki_tbl{'QOS'}; #
              $flags = $wiki_tbl{'FLAGS'};
              $accountName = $wiki_tbl{'ACCOUNT'};
              $executable = $wiki_tbl{'COMMAND'}; 
              $extensions = $wiki_tbl{'EXTENTIONS'}; ##
              $bypassCount = $wiki_tbl{'BYPASSCOUNT'}; #
              $PSUtil = $wiki_tbl{'PSUTILIZED'}; #
              $partitionName = $wiki_tbl{'PARTITION'};
              # procsPerTask is not in WIKI
              $memoryPerTask = $wiki_tbl{'DMEM'}; #
              $diskPerTask = $wiki_tbl{'DDISK'}; #
              $swapPerTask = $wiki_tbl{'DSWAP'}; #
              $activationTime = $wiki_tbl{'ACTIVATIONTIME'}; ##
              $deadlineTime = $wiki_tbl{'DEADLINETIMET'}; ##
              $allocatedHostlist = $wiki_tbl{'TASKMAP'};
              $resourceManager = $wiki_tbl{'RM'}; #
              $requestedHostList = $wiki_tbl{'HOSTLIST'};
              $reservationId = $wiki_tbl{'RESERVATION'}; #
              $simData = $wiki_tbl{'SIM'}; ##
              $nodeConstraints = $wiki_tbl{'NODEALLOCATIONPOLICY'}; #
              $messages = $wiki_tbl{'MESSAGES'};
              $cost = $wiki_tbl{'COST'}; #
              $history = $wiki_tbl{'HISTORY'}; ## 
              $utilization = $wiki_tbl{'PSUTILIZED'}; ##
              $estimateData = $wiki_tbl{'ESTIMATEDATA'}; ##
              $complCodeIWDOut = $wiki_tbl{'IWD'}; ##
              $memoryStats = $wiki_tbl{'JOBMEM'}; ##
              $cpuStats = $wiki_tbl{'PSUTILIZED'}; #
              
              # WIKI memory is unitless
              $memory .= 'M';
              
              # need for node count
              if ($allocatedHostlist)
              {
                my @taskArr = split /,/, $allocatedHostlist;
                $nodes = scalar @taskArr;             
              }
              
              @fields = ();
              
              @fields = 
              ($hRTime,         $timestamp,       $objectType,
               $objectId,       $objectEvent,     $nodes,
               $tasks,          $userName,        $groupName,
               $wallclockLimit, $completionState, $className,
               $submissionTime, $dispatchTime,    $startTime,
               $completionTime, $architecture,    $opsys,
               $memcmp,         $memory,          $diskcmp,
               $disk,           $features,        $systemQueueTime,
               $allocatedTasks, $TPN,             $qosName,
               $flags,          $accountName);

              my @restOfFields = (
                                  $executable,
                                  $extra,
                                  $extensions,
                                  $bypassCount,
                                  $PSUtil,
                                  $partitionName,
                                  $memoryPerTask,
                                  $diskPerTask,
                                  $swapPerTask,
                                  $activationTime,
                                  $deadlineTime,
                                  $allocatedHostlist,
                                  $resourceManager,
                                  $requestedHostList,
                                  $reservationId,
                                  $simData,
                                  $nodeConstraints,
                                  $messages,
                                  $cost,
                                  $history,
                                  $utilization,
                                  $estimateData,
                                  $complCodeIWDOut,
                                  $memoryStats,
                                  $cpuStats,
                                 );
              
              @fields = (@fields , @restOfFields);
            }                      
            
            # Filter results
            if ($object) { next unless $objectType =~ /$object/i; }
            if ($job)    { next unless $objectType =~ /job/i; }
            if ($id)     { next unless $objectId   =~ /^$id$/i; }
            if ($account)
            {
                next
                  unless defined $accountName
                  && $accountName =~ /$account/i;
            }
            if ($class)
            {
                next unless defined $className && $className =~ /$class/i;
            }
            if ($group)
            {
                next unless defined $groupName && $groupName =~ /$group/i;
            }
            if ($qos)
            {
                next unless defined $qosName && $qosName =~ /$qos/i;
            }
            if ($user)
            {
                next unless defined $userName && $userName =~ /$user/i;
            }

            # Parse the event
            my %event = ();
            foreach my $field (@show)
            {
              # WIKI does not have - for no value
              $fields[$fieldIndex{$field}] = '-'
              if not defined $fields[$fieldIndex{$field}];
              
              my $formattedField;
              if ($fields[$fieldIndex{$field}] ne '-'
                  and $field =~ /Time/)
              {
                $formattedField = 
                localtime $fields[$fieldIndex{$field}] 
                if defined $fields[$fieldIndex{$field}];
                
              }
              elsif ($field =~ /Details/)
              {
                $formattedField = $line;
              }
              else
              {
                $formattedField = $fields[$fieldIndex{$field}];
              }
              $event{$field}     = $formattedField;
              
              # Uninitialized var error
              if (defined $formattedField) 
              {
                $fieldSize{$field} = length $formattedField
                if length $formattedField > $fieldSize{$field};
              }

            }
            push @events, \%event;
        }
    }

    # Print the header
    my @header = ();
    foreach my $field (@show)
    {
        push @header, sprintf "%-$fieldSize{$field}s", $field;
    }
    print join(' ', @header), "\n";

    # Print the border
    my @border = ();
    foreach my $field (@show)
    {
        push @border, sprintf "%-$fieldSize{$field}s", '-' x $fieldSize{$field};
    }
    print join(' ', @border), "\n";

    # Print the event
    foreach my $eventRef (@events)
    {
        my @event = ();
        foreach my $field (@show)
        {
            # Unintialized if empty
            push @event, sprintf "%-$fieldSize{$field}s", $eventRef->{$field} if defined $eventRef->{$field};
          
        }
        
        print join(' ', @event), "\n";
    }

    # Exit with status code
    exit 0;
}

################################################################################
# Sorts events files by epoch time
################################################################################
sub by_date
{
    return epochDate($a) <=> epochDate($b);
}

################################################################################
# $epochDate = epochDate($eventFile)
# Returns the epoch date for an event file
################################################################################
sub epochDate
{
    my ($eventFile) = @_;

    $eventFile =~ s/.*\///;
    my ($basename, $wday, $mon, $day, $year) = split /[\._]/, $eventFile;
    my $epochDate = timelocal(0, 0, 0, $day, $monthToIndex{$mon}, $year - 1900);

    return $epochDate;
}

##############################################################################

__END__

=head1 NAME

B<showevents> - list historical moab events

=head1 SYNOPSIS

B<showevents> [B<-a> I<account_name>] [B<-c> I<class_name>] [B<-g> I<group_name>] [B<-j>] [B<-o> I<object_type>] [B<-n> I<days>] [B<-q> I<qos_name>] [B<-u> I<user_name>] [B<--show> I<attribute_name>[,I<attribute_name>...]] [B<--help>] [B<--man>] [[B<-i>] <object id>]

=head1 DESCRIPTION

The B<showevents> command is used to list past moab event records. It searches through the specified number of events files (or all if number not specified) while filtering according to the specified options.

The event time, object type, object id, event type and event data are displayed for each selected event.

=head1 OPTIONS

=over 4

=item B<-a> I<account_name>

Show only event records matching the specified account.

=item B<-c> I<class_name>

Show only event records matching the specified class (queue).

=item B<-g> I<group_name>

Show only event records matching the specified group.

=item B<-j>

Show only job event records. Same as B<-o> job.

=item B<-n> I<days>

Restricts the number of past events files to search.

=item B<-o> I<object_type>

Show only event records of the specified object type.

=item B<-q> I<qos_name>

Show only event records matching the specified qos.

=item B<-u> I<user_name>

Show only event records matching the specified user.

=item [B<-i>] I<object_id>

Show only event records matching the specified object id (such as a job id).

=item B<--show> I<attribute_name>[,I<attribute_name>]...

Displays only the specified attributes in the order listed. Valid attributes include: Account, AllocatedHostList, AllocatedTasks, Arch, Bypass, Class, Details, Disk, DispatchTime, Status, EndTime, Event, EventTime, Executable, Extra, Features, Flags, Group, Id, Memory, Message, Nodes, Object, Opsys, Partition, Procs, ProcsPerNode, PSUtil, QOS, RequiredHostList, Reservation, RM, StartTime, SubmitTime, User, Utilization, WallclockLimit

=item B<-? | --help>

brief help message

=item B<--man>

full documentation

=back

=head1 EXAMPLE

Show all events involving job id 220 and restrict the search to the last 4 days.

showevents -n 4 -t job 220

=head1 SEE ALSO

Additional information can be found at http://www.adaptivecomputing.com/

=head1 AUTHOR

Adaptive Computing, Inc., E<lt>http://www.adaptivecomputing.com/E<gt>

=head1 COPYRIGHT AND LICENSE

Copyright (C) 2006-2016 by Adaptive Computing Enterprises, Inc. All Rights Reserved.

=cut
