#! /usr/bin/perl -w
################################################################################
#
# showhist - list historical job information
#
#   Copyright (C) 2007-2016 by Adaptive Computing Enterprises, Inc. All Rights Reserved.
#
################################################################################

use strict;
use FindBin;
use lib "/opt/moab/lib/perl5", "/opt/moab/etc";
use lib "${FindBin::Bin}/../../lib/perl5", "${FindBin::Bin}/../../etc";
use Getopt::Long 2.24 qw(:config no_ignore_case);
use Time::Local;
use Moab::Tools qw(:DEFAULT :date);    # Must be before required config file
use autouse 'Pod::Usage' => qw(pod2usage);
BEGIN { require "config.moab.pl"; }
our ($homeDir, $logLevel);

my @names = (
    "Job Id",
    "Job Name",
    "Executable",
    "Output File",
    "Working Directory",

    "User Name",
    "Group Name",
    "Account Name",
    "Machine Name",
    "Class Name",
    "Queue Name",
    "Quality Of Service",
    "Cluster Name",
    "Partition Name",
    "Resource Manager Name",

    "Architecture",
    "Operating System",
    "Features Requested",
    "Job Flags",

    "Node Count",
    "Processor Count",
    "Procs Per Node",
    "Wallclock Limit",
    "Wallclock Duration",
    "CPU Duration",

    "Memory Requested",
    "Memory Used",
    "Disk Requested",
    "Disk Used",
    "Procs Per Task",
    "Memory Per Task",
    "Disk Per Task",
    "Swap Per Task",

    "Submit Time",
    "Start Time",
    "End Time",
    "Completion State",
    "Exit Code",
    "Activation Time",
    "Deadline Time",

    "Master Node",
    "Allocated Nodelist",
    "Requested Nodelist",

    "Reservation Id",
    "Messages",
);

my %wiki_tbl;
my @wiki_term;
my $exitCode;

logPrint("Invoked: $0 " . join(' ', @ARGV) . "\n") if $logLevel;

# Parse Command Line Arguments
my ($account, $class, $endDate, $group, $help, $man, $node, $num, $qos,
    $specifiedJobId, $startDate, $user);
GetOptions(
    'account=s'     => \$account,
    'class|queue=s' => \$class,
    'end=s'         => \$endDate,
    'group=s'       => \$group,
    'help|?'        => \$help,
    'jobid=s'       => \$specifiedJobId,
    'man'           => \$man,
    'node|host=s'   => \$node,
    'n|num=s'       => \$num,
    'q|qos=s'       => \$qos,
    'start=s'       => \$startDate,
    'user=s'        => \$user,
) or pod2usage(2);

# Display usage if necessary
pod2usage(2) if $help;
if ($man)
{
    if ($< == 0)    # Cannot invoke perldoc as root
    {
        my $id = eval { getpwnam("nobody") };
        $id = eval { getpwnam("nouser") } unless defined $id;
        $id = -2                          unless defined $id;
        $<  = $id;
    }
    $> = $<;                         # Disengage setuid
    $ENV{PATH} = "/bin:/usr/bin";    # Untaint PATH
    delete @ENV{'IFS', 'CDPATH', 'ENV', 'BASH_ENV'};
    if ($0 =~ /^([-\/\w\.]+)$/) { $0 = $1; }    # Untaint $0
    else { die "Illegal characters were found in \$0 ($0)\n"; }
    pod2usage(-exitstatus => 0, -verbose => 2);
}

# Use remaining argument as job id
if (@ARGV)
{
    if (@ARGV == 1 && !$specifiedJobId)
    {
        ($specifiedJobId) = @ARGV;
        ($specifiedJobId) =~ s/\[/\\\[/;
        ($specifiedJobId) =~ s/]/\\\]/;
    }
    else
    {
        pod2usage(2);
    }
}

# Build a sorted list of stats files
my @eventsFiles = glob("${homeDir}/stats/events.*_*_*_*");
@eventsFiles = sort files_by_date @eventsFiles;
if ($startDate)
{
    if ($startDate =~ /^(\d{4})[\/\-](\d{2})[\/\-](\d{2})$/)
    {
        my $startEpoch = timelocal(0, 0, 0, $3, $2 - 1, $1 - 1900);
        @eventsFiles = grep { epochDate($_) >= $startEpoch } @eventsFiles;
    }
    else
    {
        logDie("Start Date ($startDate) is not in YYYY-MM-DD format.\n");
    }
}
if ($endDate)
{
    if ($endDate =~ /^(\d{4})[\/\-](\d{2})[\/\-](\d{2})$/)
    {
        my $endEpoch = timelocal(0, 0, 0, $3, $2 - 1, $1 - 1900);
        @eventsFiles = grep { epochDate($_) <= $endEpoch } @eventsFiles;
    }
    else
    {
        logDie("End Date ($endDate) is not in YYYY-MM-DD format.\n");
    }
}
if ($num)
{
    my $firstIndex = ($num <= @eventsFiles) ? @eventsFiles - $num : 0;
    my $lastIndex = $#eventsFiles;
    @eventsFiles = @eventsFiles[$firstIndex .. $lastIndex];
}

# Parse the event files and populate %job
my %job = ();
foreach my $eventsFile (@eventsFiles)
{
    open EVENTS, "< $eventsFile"
      or logDie("Unable to open events file ($eventsFile) for reading: $!\n");
    foreach my $line (<EVENTS>)
    {
        chomp $line;
        my @values = split /\s+/, $line;
        my (
            $hRTime,          $timestamp,         $objectType,
            $objectId,        $objectEvent,       $nodeCount,
            $taskCount,       $userName,          $groupName,
            $wallclockLimit,  $completionState,   $className,
            $submissionTime,  $dispatchTime,      $startTime,
            $completionTime,                      $architecture,
            $opsys,           $memcmp,            $memory,
            $diskcmp,         $disk,              $features,
            $systemQueueTime, $allocatedTasks,    $TPN,
            $qosName,         $flags,             $accountName,
            $executable,      $extensions,        $bypassCount,
            $PSUtil,          $partitionName,     $procsPerTask,
            $memoryPerTask,   $diskPerTask,       $swapPerTask,
            $activationTime,  $deadlineTime,      $allocatedHostlist,
            $resourceManager, $requestedHostList, $reservationId,
            $simData,         $nodeConstraints,   $messages,
            $cost,            $history,           $utilization,
            $estimateData,    $complCodeIWDOut,   $memoryStats,
            $cpuStats,
        ) = @values;
        next unless @values >= 6;

        # Filter out irrelevant events
        next unless $objectType =~ /job/i;

        # If values are not defined in WIKI then use non-WIKI
        %wiki_tbl = ();
        undef $exitCode;

        # TODO: Ensure that this will not be in the non-WIKI format
        if ($taskCount =~ /=/)
        {
          # WIKI terms start after the header
          for my $val (@values[6..$#values])
          {
            if ($val !~ /=/)
            {
              print STDERR "WIKI term \'$val\' does not contain =\n";
              next;
            }
            
            # limit for memory compare terms
            @wiki_term = split /=/, $val, 2;
            
            $wiki_tbl{$wiki_term[0]} =  $wiki_term[1];
            
          }
          
          $wiki_tbl{'OBJID'} = $objectId;
          $taskCount = $wiki_tbl{'TASKS'};
          $userName = $wiki_tbl{'UNAME'};
          $groupName = $wiki_tbl{'GNAME'};
          $className = $wiki_tbl{'RCLASS'}; 
          $submissionTime = $wiki_tbl{'SUBMITTIME'};
          $startTime = $wiki_tbl{'STARTTIME'};
          $completionTime = $wiki_tbl{'COMPLETETIME'};
          $memory = $wiki_tbl{'DMEM'}; 
          $qosName = $wiki_tbl{'QOS'}; #
          $accountName = $wiki_tbl{'ACCOUNT'};
          $executable = $wiki_tbl{'COMMAND'}; 
          $allocatedHostlist = $wiki_tbl{'TASKMAP'};
          $exitCode = $wiki_tbl{'EXITCODE'};
          
          # node count does not exist
          
          if ($allocatedHostlist)
          {
            my @taskArr = split /,/, $allocatedHostlist;
            $nodeCount = scalar @taskArr;             
          }
        }                      
                                
        if ($specifiedJobId)
        {
            next unless $objectId =~ /^$specifiedJobId$/i;
        }
        if ($account)
        {
            next
              unless defined $accountName
              && $accountName =~ /$account/i;
        }
        if ($class)
        {
            next unless defined $className && $className =~ /$class/i;
        }
        if ($group)
        {
            next unless defined $groupName && $groupName =~ /$group/i;
        }
        if ($node)
        {
            next
              unless defined $allocatedHostlist
              && $allocatedHostlist =~ /$node/i;
        }
        if ($qos)
        {
            next unless defined $qosName && $qosName =~ /$qos/i;
        }
        if ($user)
        {
            next unless defined $userName && $userName =~ /$user/i;
        }

        # Parse the event
        my $jobId = $objectId;
        if ($userName) { $job{$jobId}{'User Name'} = $userName; }
        if (defined $groupName && $groupName ne '-')
        {
            $job{$jobId}{'Group Name'} = $groupName;
        }
        if (defined $accountName && $accountName ne '-')
        {
            $job{$jobId}{'Account Name'} = $accountName;
        }
        if (defined $className && $className ne '-')
        {
            $className = $1 if $className =~ /(\w+)/;
            $job{$jobId}{'Queue Name'} = $className;
        }
        if (defined $qosName && $qosName ne '-:-')
        {
            $qosName = $1 if $qosName =~ /(.*):/;
            $job{$jobId}{'Quality Of Service'} = $qosName;
        }
        if ($submissionTime)
        {
            $job{$jobId}{'Submit Time'} = localtime $submissionTime;
        }
        if ($startTime)
        {
            $job{$jobId}{'Start Time'} = localtime $startTime;
            if ($completionTime)
            {
                $job{$jobId}{'End Epoch'} = $completionTime;
                $job{$jobId}{'End Time'}  = localtime $completionTime;
                $job{$jobId}{'Wallclock Duration'} =
                  ddhhmmss($completionTime - $startTime);
                #$job{$jobId}{'CPU Duration'} = ddhhmmss($PSUtil);
            }
        }

        # WIKI has exit code
        $exitCode = $wiki_tbl{'EXITCODE'};
        $job{$jobId}{'Exit Code'} = $exitCode if
        defined $exitCode;
        
        if (not defined $exitCode and $complCodeIWDOut and $complCodeIWDOut ne '-')
        {
            chomp $complCodeIWDOut;
            my @splitfield = split /,/, $complCodeIWDOut;
            $job{$jobId}{'Exit Code'} = $splitfield[0];
        }
        if ($complCodeIWDOut and $complCodeIWDOut ne '-')
        {
          chomp $complCodeIWDOut;
          my @splitfield   = split /,/, $complCodeIWDOut;
          my ($outDir)     = grep(/^OutputFile/, @splitfield);
          my ($workingDir) = grep(/^IWD/,        @splitfield);
          $workingDir =~ s/IWD=(.*)/$1/;
          $job{$jobId}{'Working Directory'} = $workingDir
            if defined $workingDir
                   and $workingDir;

          if (defined $outDir)
          {
            $outDir =~ s/OutputFile=(.*)/$1/;
            if ($outDir =~ m/^\/.*/)
            {
              $job{$jobId}{'Output File'} = $outDir;
            }
            else
            {
              $job{$jobId}{'Output File'} = $workingDir . '/' . $outDir
                if defined $workingDir;
            }
          }
        }
        if ($executable and $executable ne '-')
        {
            $job{$jobId}{'Executable'} = $executable;
        }
        if ($nodeCount)
        {
            $job{$jobId}{'Node Count'} = $nodeCount;
        }
        if ($taskCount and not defined $wiki_tbl{'TASKS'})
        {
          my $procCount = $taskCount;
          if ($procsPerTask)
          {
            $procCount *= $procsPerTask
            if defined $procsPerTask;
          }
          $job{$jobId}{'Processor Count'} = $procCount;
          # WIKI has DPROCS
        }

        # WIKI has DPROCS
        $job{$jobId}{'Processor Count'} = $wiki_tbl{'DPROCS'}
          if defined $wiki_tbl{'DPROCS'};
          
        # This is memory requested
        #if ($memory)
        #{
        #    $job{$jobId}{'Memory Used'} = $memory;
        #}
        if ($allocatedHostlist)
        {
            my %allocatedNode = ();
            if ($allocatedHostlist =~ /\[/)
            {
                map { $_ =~ s/\*\d+//, $allocatedNode{$_}++ } split /:/,
                  $allocatedHostlist;
            }
            else
            {
                map { $allocatedNode{$_}++ } split /,/, $allocatedHostlist;
            }
            $job{$jobId}{'Allocated Nodelist'} = join ':',
              sort keys %allocatedNode;
          
        }
    }
}

# Print out the jobs
foreach my $jobId (sort jobs_by_date keys %job)
{
    next unless $job{$jobId}{'End Epoch'};
    printf "%-18s: %s\n", "Job Id", $jobId;
    foreach my $name (@names)
    {
        if (defined $job{$jobId}{$name})
        {
            printf "%-18s: %s\n", $name, $job{$jobId}{$name};
        }
    }
    # Print job script (last)
    if (-e "$homeDir/stats/jobarchive/$jobId")
    {
        my $errorOpenFile = 0;
        if (open(JOBSCR, "$homeDir/stats/jobarchive/$jobId"))
        {
            print "#### Start Job Batch Script ####\n";
            while (<JOBSCR>)
            {
                print;
            }
            close(JOBSCR);
            print "#### End Job Batch Script ####\n";
        }
    }
    print "\n";
}

# Exit with status code
exit 0;

################################################################################
# Sorts events files by epoch time
################################################################################
sub files_by_date
{
    return epochDate($a) <=> epochDate($b);
}

################################################################################
# Sorts jobs by (epoch time) end date
################################################################################
sub jobs_by_date
{
    if ($job{$a}{'End Epoch'} && $job{$b}{'End Epoch'})
    {
        return $job{$a}{'End Epoch'} <=> $job{$b}{'End Epoch'};
    }
    elsif ($job{$a}{'End Epoch'}) { return -1; }
    elsif ($job{$b}{'End Epoch'}) { return 1; }
    else                          { return $a cmp $b; }
}

################################################################################
# $epochDate = epochDate($eventFile)
# Returns the epoch date for an event file
################################################################################
sub epochDate
{
    my ($eventFile) = @_;

    $eventFile =~ s/.*\///;
    my ($basename, $wday, $mon, $day, $year) = split /[\._]/, $eventFile;
    my $epochDate = timelocal(0, 0, 0, $day, $monthToIndex{$mon}, $year - 1900);

    return $epochDate;
}

################################################################################
# $ddhhmmss = ddhhmmss($seconds)
# Converts a duration in seconds to ddhhmm
################################################################################
sub ddhhmmss
{
    my ($seconds) = @_;

    # Convert duration in minutes to ddhhmm
    my $days = int($seconds / 86400);
    $seconds = $seconds % 86400;
    my $hours = int($seconds / 3600);
    $seconds = $seconds % 3600;
    my $minutes = int($seconds / 60);
    $seconds = $seconds % 60;

    if ($days)
    {
        return sprintf('%d:%02d:%02d:%02d', $days, $hours, $minutes, $seconds);
    }
    else { return sprintf('%02d:%02d:%02d', $hours, $minutes, $seconds); }
}

##############################################################################

__END__

=head1 NAME

B<showhist> - list historical job information

=head1 SYNOPSIS

B<showhist> [B<-a> I<account_name>] [B<-c>|B<-queue> I<class_name>] [B<-g> I<group_name>] [B<-q> I<qos_name>] [B<-u> I<user_name>] [B<-n> I<days>] [B<-s> I<start date>] [B<-e> I<end date>] [B<--help>] [B<--man>] [[B<-j>] <job id>]

=head1 DESCRIPTION

The B<showhist> command is used to list past job information. It searches through the designated events files while filtering according to the specified options. The relevant fields for each job are shown in a multi-line format, with a blank line between jobs.

=head1 OPTIONS

=over 4

=item B<-a> I<account_name>

Show only event records matching the specified account.

=item B<-c> I<class_name>

Show only event records matching the specified class (queue).

=item B<-e> I<end_date>

Restricts the search to event files ending with the specified date. The date is specified in the format YYYY-MM-DD. The default query searches to the latest available event file.

=item B<-g> I<group_name>

Show only event records matching the specified group.

=item [B<-j>] I<job_id>

Show only event records matching the specified job id.

=item B<-n|--num> I<days>

Restricts the number of past events files to search.

=item B<-q> I<qos_name>

Show only event records matching the specified qos.

=item B<-s> I<start_date>

Restricts the search to event files starting with the specified date. The date is specified in the format YYYY-MM-DD. The default query searches from the earliest available event file.

Restricts the search to event files starting with the specified date.

=item B<-u> I<user_name>

Show only event records matching the specified user.

=item B<-? | --help>

brief help message

=item B<--man>

full documentation

=back

=head1 EXAMPLE

Show job information for job id 220 and restrict the search to the last 4 days.

showhist.moab.pl -n 4 -j 220

=head1 SEE ALSO

Additional information can be found at http://www.adaptivecomputing.com/

=head1 AUTHOR

Adaptive Computing, Inc., E<lt>http://www.adaptivecomputing.com/E<gt>

=head1 COPYRIGHT AND LICENSE

Copyright (C) 2007-2016 by Adaptive Computing Enterprises, Inc. All Rights Reserved.

=cut
