#!/usr/bin/env python
# ====================================================================
# support-dig script that gathers diagnostics for Adaptive Computing 
# support. 
#
# Copyright (C) 2005-2016 by Adaptive Computing Enterprises, Inc. 
# All Rights Reserved.
#
# THIS SOFTWARE SCRIPT IS PROVIDED BY ADAPTIVE COMPUTING ENTERPRISES, 
# INC. "AS IS" AND IS FOR REFERENCE USE ONLY.  ANY EXPRESS OR IMPLIED 
# WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
# DISCLAIMED. IN NO EVENT SHALL ADAPTIVE COMPUTING ENTERPRISES, INC. 
# BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, 
# OR CONSEQUENTIAL DAMAGES ARISING FROM THE USE THEREOF.
# ====================================================================

# ====================================================================
# DEVELOPER NOTES:
# Although support-diag was written against 2.6.6 you could run it 
# against 2.4. Simply modify the python_minor_version logic on line 32
# e.g. if python_major_version >= 2 and python_minor_version < 4:
# ====================================================================

# ====================================================================
# Import from libraries and validate Python version
# ====================================================================
import sys
import os

# Validate we have the correct Python version
python_major_version = sys.version_info[0]
python_minor_version = sys.version_info[1]
if python_major_version >= 2 and python_minor_version < 6:
    print "This script requires Python version 2.6 or greater and you have %s.%s" % \
          (python_major_version, python_minor_version)
    sys.exit(1)

if os.geteuid() != 0:
    sys.exit("You need to have root privileges to run this script.\n "
             "Please try again, this time using 'sudo'. Exiting.")

# Import libraries
import re
import subprocess
import shutil
import glob
import threading
import time
import datetime
import tarfile
from optparse import OptionParser

# Import AC specifc libraries
from lib.ACcray import *
from lib.ACmam import *
from lib.ACmoab import *
from lib.ACmws import *
from lib.ACslurm import *
from lib.ACtorque import *
from lib.ACviewpoint import *
from lib.ACos import *
import lib.ACgeneral as ACgeneral

# ====================================================================
# Modified variables
# ====================================================================
# Version number
version_number = "2.0"

# Time
today_date_time = datetime.datetime.now()
today_date_time = ACgeneral.format_iso_date(today_date_time)

# File Transfer
use_ssh = True
use_ftp = False

tmp_directory = "/tmp/"
output_directory = "/tmp/"
diag_date = ""
hostname = ""
current_directory = ""
ticket_number = "1111"
moab_timeout = 300
core_dumps = 0

# ====================================================================
# Command line option parsing
# ====================================================================
parser = OptionParser()

parser.add_option("-p", "--tmp-dir", dest="tmp_dir",
                  help="Use a different tmp directory to store output.", metavar="TMPDIR")
parser.add_option("-n", "--no-upload", dest="no_upload",
                  help="Do not upload to Adaptive Computing", default=False, action="store_true")
parser.add_option("-t", "--ticket-number", dest="ticket_number",
                  help="Support ticket number", metavar="TICKET#")
parser.add_option("-u", "--moab-timeout", dest="moab_time_out",
                  help="Define moab command timeout (Default 300 seconds).", metavar="TIMEOUT")
parser.add_option("-o", "--offline-mode", dest="offline_mode",
                  help="Gather logging only.", default=False, action="store_true")
parser.add_option("-r", "--ftp", dest="ftp_mode",
                  help="Use ftp instead of scp.", default=False, action="store_true")
parser.add_option("-d", "--debug", dest="debug_mode",
                  help="Print support-diag debugging information.", default=False, action="store_true")
parser.add_option("-v", "--no-vp", "--without-viewpoint", dest="diag_viewpoint",
                  help="Disable viewpoint diagnostics.", default=False, action="store_true")
parser.add_option("-w", "--no-mws", "--without-mws", dest="diag_mws",
                  help="Disable mws diagnostics.", default=False, action="store_true")
parser.add_option("-a", "--no-mam", "--without-mam", dest="diag_mam",
                  help="Disable mam diagnostics.", default=False, action="store_true")
parser.add_option("-x", "--no-os", "--without-os", dest="diag_os",
                  help="Disable os diagnostics.", default=False, action="store_true")
parser.add_option("-c", "--no-cray", "--without-cray", dest="diag_cray",
                  help="Disable Cray diagnostics.", default=False, action="store_true")
parser.add_option("-s", "--no-slurm", "--without-slurm", dest="diag_slurm",
                  help="Disable SLURM diagnostics.", default=False, action="store_true")
parser.add_option("-m", "--no-moab", "--without-moab", dest="diag_moab",
                  help="Disable Moab diagnostics.", default=False, action="store_true")
parser.add_option("-q", "--no-torque", "--without-torque", dest="diag_torque",
                  help="Disable TORQUE diagnostics.", default=False, action="store_true")
parser.add_option("-f", "--full-logging", dest="full_logging",
                  help="Gather all logging regardless of log size!", default=False, action="store_true")
parser.add_option("-y", "--gather-core", dest="gather_core",
                  help="Gather cores in the last [1|2..n|A] days. E.g -y 1 (last day), "
                       "or -y A (all).", metavar="[1|2..|A]")
parser.add_option("-V", "--version", dest="version_mode",
                  help="Print version information.", default=False, action="store_true")
(options, args) = parser.parse_args()
if options.version_mode:
    print "Support-diag version: " + version_number
    print("Copyright (C) 2005-2016 by Adaptive Computing Enterprises, Inc.")
    print("Report any bugs to Adaptive Computing via the support ticketing system.")
    sys.exit()
if options.full_logging:
    print("\n\n ***** WARNING!!!!! *****\n\n")
    print("You have requested to gather extensive logging.")
    print("This script attempts to limit the amount of data")
    print("captured. However, requesting full logging will")
    print("indiscriminately grab logging with no regard to size! \n")
    grab_all_logs = True 
    while grab_all_logs:
        remove_answer = str(raw_input("Continue (Y/N)? :"))
        remove_answer = remove_answer.upper()
        if remove_answer == "Y":
            print("You have been warned!")
            grab_all_logs = False
        if remove_answer == "N":
            sys.exit(0)

if options.ticket_number is None and options.no_upload is False:
    parser.print_help()
    print("*******************************************************")
    print("You must specify a ticket -t or -n to not upload.")
    print("*******************************************************")
    sys.exit()
# Assign tmp_dir
if options.tmp_dir is not None:
    tmp_directory = options.tmp_dir
    tmp_directory = tmp_directory.strip()
    last_car_tmp_dir = tmp_directory[-1:]
    if last_car_tmp_dir != "/":
        tmp_directory += "/"
    if not os.path.isdir(tmp_directory):
        print "The tmp directory does not exist:", tmp_directory
        sys.exit()
# Assign ticket number
if options.ticket_number is not None:
    ticket_number = options.ticket_number
# Noupload
if options.no_upload:
    no_upload = options.no_upload
# Change timeout for Moab commands
if options.moab_time_out is not None:
    # moab_timeout = " --timeout="
    moab_timeout = int(options.moab_time_out)
if options.ftp_mode:
    use_ssh = False
    use_ftp = True

# Gather core dumps
if options.gather_core:
    core_dumps = str(options.gather_core)
    core_dumps = core_dumps.upper()

# ====================================================================
# Create output dir and initialize variables
# ====================================================================
def init(tmp_directory):
    """

    :param tmp_directory: Take a temp directory
    :return:  diag_date, hostname, current_directory, output_directory
    """
    print "*** Please do NOT restart your HPC services while this script is running. ***"

    # Determine hostname
    try:
        return_code, hostname = ACgeneral.run_and_return_output("hostname")
    except:
        print("I was unable to gather the hostname: " + str(return_code))
    if return_code == 0 and hostname != '':
        hostname.rstrip('\n')
    else:
        hostname = "undefined"
        hostname.rstrip()

    # construct hostname.yyyy-mm-dd_hh-mm-ss.ticket
    time_now = datetime.datetime.now()
    diag_date = "{0}.{1}-{2}-{3}_{4}-{5}-{6}.{7}".format(hostname.rstrip(), str(time_now.year), str(time_now.month),
                                                         str(time_now.day), str(time_now.hour), str(time_now.minute),
                                                         str(time_now.second), str(ticket_number))
    # Save current directory
    current_directory = os.getcwd()
    # Make sure the tmp dir based off hostname.yyyy-mm-dd_hh-mm-ss.ticket
    diag_date = diag_date.strip()
    output_directory = tmp_directory + diag_date.strip()
    output_directory += "/"
    if not os.path.exists(output_directory):
        try:
            os.makedirs(output_directory)
        except:
            print("I was unable to make the output direcotry: " + output_directory)
            sys.exit(1)
    else:
        print "I was not able to create an output directory to write to.\n"
        print "Please ensure this script can write to the tmp path. (" + output_directory + ")"
        sys.exit()
    # Output will go to tmpdir/hostname.timestamp. This removes
    # the old formatting of var/spool... etc...
    return diag_date, hostname, current_directory, output_directory

# ====================================================================
# Main entry into the diagnostics
# ====================================================================
def main():
     
    # Var to hold diagnostic output from modules.
    script_diag_output = ""
    
    diag_date, hostname, current_directory, output_directory = init(tmp_directory)
    print ("Writing diagnostics to: " + output_directory)
    # Moab diags
    if not options.diag_moab:
        try:
            moab_o = ACmoab(today_date_time, options.full_logging, core_dumps)
            moab_o.init(output_directory, moab_timeout)
            if options.offline_mode:
                moab_o.offline_mode()
            else:
                moab_o.run()
        except:
            e = sys.exc_info()[0]
            print("There was a failure with the Moab module: %s")
        script_diag_output = moab_o.diag()
    # TORQUE diags
    if not options.diag_torque:
        try:
            torque_o = ACtorque(today_date_time, options.full_logging, core_dumps)
            torque_o.init(output_directory)
            if options.offline_mode:
                torque_o.offline_mode()
            else:
                torque_o.run()
        except:
            e = sys.exc_info()[0]
            print("There was a failure with the TORQUE module: %s")
        script_diag_output += torque_o.diag()
    # Viewpoint
    if not options.diag_viewpoint:
        try:
            viewpoint_o = ACviewpoint()
            viewpoint_o.init(output_directory)
            viewpoint_o.run()
        except:
            print("There was a failure with the viewpoint module: %s")
        script_diag_output += viewpoint_o.diag()
    # MAM
    if not options.diag_mam:
        try:
            mam_o = ACmam()
            mam_o.init(output_directory)
            if options.offline_mode:
                mam_o.offline_mode()
            else:
                mam_o.run()
        except:
            e = sys.exc_info()[0]
            print("There was a failure with the mam module: %s")
        script_diag_output += mam_o.diag()
    # MWS
    if not options.diag_mws:
        try:
            mws_o = ACmws()
            mws_o.init(output_directory)
            mws_o.run()
        except:
            e = sys.exc_info()[0]
            print("There was a failure with the mws module: %s")
        script_diag_output += mws_o.diag()
    # OS
    if not options.diag_os:
        try:
            os_o = ACos()
            os_o.init(output_directory)
            os_o.run()
        except:
            e = sys.exc_info()[0]
            print("There was a failure with the os module: %s")
        script_diag_output += os_o.diag()
    # Cray
    if not options.diag_cray:
        try:
            cray_o = ACcray()
            cray_o.init(output_directory)
            cray_o.run()
            script_diag_output += cray_o.diag()
        except:
            e = sys.exc_info()[0]
            print("There was a failure with the Cray module: %s")
    # SLURM
    if not options.diag_slurm:
        try:
            slurm_o = ACslurm(options.full_logging)
            slurm_o.init(output_directory)
            if options.offline_mode:
                slurm_o.offline_mode()
            else:
                slurm_o.run()
        except:
            e = sys.exc_info()[0]
            print("There was a failure with the SLURM module: %s")
        script_diag_output += slurm_o.diag()


    #Diagnostics
    diag_f = open(output_directory+'/script_diagnostics.txt','w')
    # Gloabals
    script_diag_output += "\n********* GLOBAL SETTINGS *********"
    script_diag_output += "\nGlobal tmp: " + tmp_directory
    script_diag_output += "Global Output dir: " + output_directory 
    script_diag_output += "\nDiag date: " + str(diag_date)
    script_diag_output += "\nHostname: " + hostname 
    diag_f.write(script_diag_output)
    diag_f.close()
    if options.debug_mode:
        print script_diag_output

    """
    Check threads before exiting and
    only allow thread to stick around for a designated timeout
    plus a delta.
    """
    thread_still_running = 0
    thread_tracking_keep_time = 0
    print("\n")
    while thread_still_running == 0:
        sys.stdout.write('\033[1A')
        sys.stdout.write("\r Threads still active:" + str(threading.active_count()) +
                         " timeout in 'seconds': (" + str((moab_timeout+10) - thread_tracking_keep_time) + ")\n")
        sys.stdout.flush()
        if threading.active_count() > 1:
            thread_still_running = 0
        else:
            thread_still_running = 1
        time.sleep(1)
        if (moab_timeout + 10) == thread_tracking_keep_time:
            thread_still_running = 1
        thread_tracking_keep_time = thread_tracking_keep_time + 1
    try:    
        ACgeneral.bundle_package(diag_date, output_directory, tmp_directory)
    except:
        print("I was unable to bundle the package.")
        ACgeneral.clean_up(output_directory)
        sys.exit(1)
    if not options.no_upload:
        if not options.ftp_mode:
            ACgeneral.transfer_package(diag_date, tmp_directory, "ssh")
        else:
            ACgeneral.transfer_package(diag_date, tmp_directory, "ftp")
    # Clean up output directory
    try:
        ACgeneral.clean_up(output_directory)
    except:
        print("I was unable to remove the dir: " + output_directory)
if __name__ == '__main__':
    main()
