#!/usr/bin/python

import sys
import multiprocessing
import subprocess
import argparse
import json

parser = argparse.ArgumentParser(description="Node power script used by the MWS power management plugin for IPMI nodes.")
parser.add_argument("moab_node_list",
    nargs="+",
    help="One or more moab node names.")
parser.add_argument("power_state",
    choices=["on","off","wake"],
    help="This translates into the 'chassis power' command that ipmitool provides with the values of on, off, and soft respectively.")
parser.add_argument("-u", "--ipmi_username_file",
    default="/opt/moab/etc/power-management/sample-power-script-username",
    help="File containing the ipmi username(If not supplied, use '/opt/moab/etc/power-management/sample-power-script-username'.)")
parser.add_argument("-p", "--ipmi_password_file",
    default="/opt/moab/etc/power-management/sample-power-script-password",
    help="File containing the ipmi password (If not supplied, use '/opt/moab/etc/power-management/sample-power-script-password'.)")
parser.add_argument("-m", "--ipmi_mapping_file",
    default="/opt/moab/etc/power-management/node-bmc-map",
    help="File containing the mappings between the moab node names and the ipmi addresses (If not supplied, use '/opt/moab/etc/power-management/node-bmc-map'.)")
parser.add_argument("-t", "--thread_count",
    default=4,
    type=int,
    help="Thread pool size that issues the IPMI commands. If not supplied, use 4")
args = parser.parse_args()
ipmi_username = None
with open (args.ipmi_username_file, "r") as file:
    for line in file.readlines():
      line.strip()
      if line[0] == '#':
        continue
      ipmi_username=line.rstrip()

if (ipmi_username is None) or (len(ipmi_username) == 0):
    ipmi_username = '""'

def runIpmiToolCommand(node):
  powercmd = "ipmitool -I lan -H %s -U %s -f %s chassis power %s" %  \
    (node[1], ipmi_username, args.ipmi_password_file, \
    'soft' if args.power_state == 'wake' else args.power_state)
  p = subprocess.Popen(powercmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
  exit_code = p.wait()
  results = p.stdout.readlines()
  node_map = { "node": node[0]}
  if exit_code:
    node_map["command"] = powercmd
    node_map["error"] = results
  else:
    node_map["output"] = results
  return node_map

entries = {}

try:
  f = open(args.ipmi_mapping_file, 'r') 
except IOError:
  sys.stderr.write("ERROR: Unable to read ipmi mappings file %s\n" % args.ipmi_mapping_file)
  sys.exit(1)

for line in f.readlines():
  line.strip()
  if line[0] == '#':
    continue
  entry = line.split()
  if len(entry) < 2:
    continue
  entries[entry[0]] = entry[1]

nodelist = []
for node in args.moab_node_list:
  bmchost = entries.get(node)
  if (bmchost == None):
    sys.stderr.write("Requested node %s does not have an associated ipmi address.  Add to %s.\n" % (node, args.ipmi_mapping_file))
    continue
  nodelist.append((node,bmchost)) 

if not nodelist:
    sys.stderr.write("All the requested nodes do not have an associated ipmi address.  Nothing to do.\n")
    exit(1);

if __name__ == '__main__':
    pool = multiprocessing.Pool(args.thread_count)
    results = []
    r = pool.map_async(runIpmiToolCommand, nodelist, callback=results.append)
    r.wait()
    if not results:
      sys.stderr.write("Error doing queries to nodes: %s\n" % args.moab_node_list)
      exit(1);
    print json.dumps(results[0])
