#!/usr/bin/python
import re

import sys
import multiprocessing
import subprocess
import argparse
import json

parser = argparse.ArgumentParser(description="Node query script used by the MWS power management plugin for IPMI nodes. This prints to standard out a list of JSON objects corresponding to each node in the node_list with at least the node name and power state according to ipmi.  An example would look like this: [{'name':'node01','power':'on','key1': 'value1','key2': 'value2' }] The rest of the key value pairs will be reported to moab as generic resources for that node.")
parser.add_argument("moab_node_list",
    nargs="+",
    help="One or more moab node names.")
parser.add_argument("-u", "--ipmi_username_file",
    default="/opt/moab/etc/power-management/sample-power-script-username",
    help="File containing the ipmi username(If not supplied, use '/opt/moab/etc/power-management/sample-power-script-username'.)")
parser.add_argument("-p", "--ipmi_password_file",
    default="/opt/moab/etc/power-management/sample-power-script-password",
    help="File containing the ipmi password (If not supplied, use '/opt/moab/etc/power-management/sample-power-script-password'.)")
parser.add_argument("-m", "--ipmi_mapping_file",
    default="/opt/moab/etc/power-management/node-bmc-map",
    help="File containing the mappings between the moab node names and the ipmi addresses (If not supplied, use '/opt/moab/etc/power-management/node-bmc-map'.)")
parser.add_argument("-t", "--thread_count",
    default=4,
    type=int,
    help="Thread pool size that issues the IPMI commands. If not supplied, use 4")
args = parser.parse_args()
ipmi_username = None
with open (args.ipmi_username_file, "r") as file:
    for line in file.readlines():
      line.strip()
      if line[0] == '#':
        continue
      ipmi_username=line.rstrip()

if (ipmi_username is None) or (len(ipmi_username) == 0):
    ipmi_username = '""'

def runIpmiToolCommand(node):
  power_query = "ipmitool -I lan -H %s -U %s -f %s chassis power status" % (node[1], ipmi_username, args.ipmi_password_file)
  p = subprocess.Popen(power_query, shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
  p.wait()
  results = p.stdout.readlines()

  match = re.match(r'.*Power is (.*).*', "".join(results), re.M | re.I)
  if not match:
    return { "name": node[0], "error":results,"command":power_query}
  power_status = match.group(1)
  if not power_status:
    return { "name": node[0], "error":results,"command":power_query}
  node_map = { "name": node[0]}
  if "on" in power_status.lower():
    node_map["power"] =  "ON"
  elif "off" in power_status.lower():
    node_map["power"] =  "OFF"
  else:
    return { "name": node[0], "error":results,"command":power_query}
  temp_query = "ipmitool -I lan -H %s -U %s -f %s sdr type temperature" % (node[1], ipmi_username, args.ipmi_password_file)
  p = subprocess.Popen(temp_query, shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
  p.wait()
  results = p.stdout.readlines()
  for result in results:
      if result.find("Temp") != -1 and result.find("degrees C") != -1:
          values = result.split('|')
          values[0] = values[0].strip()                 #Strip all trailing spaces.
          values[0] = values[0].replace(' ','_')        #Replace space in "System Temp" to "System_Temp"
          node_map[values[0]] =  int(values[4].split()[0])
  return node_map

entries = {}
try:
  f = open(args.ipmi_mapping_file, 'r') 
except IOError:
  sys.stderr.write("ERROR: Unable to read ipmi mappings file %s\n" % args.ipmi_mapping_file)
  sys.exit(1)

for line in f.readlines():
  line.strip()
  if line[0] == '#':
    continue
  entry = line.split()
  if len(entry) < 2:
    continue
  entries[entry[0]] = entry[1]

nodelist = []
for node in args.moab_node_list:
  bmchost = entries.get(node)
  if (bmchost == None):
    sys.stderr.write("Requested node %s does not have an associated ipmi address.  Add to %s.\n" % (node, args.ipmi_mapping_file))
    continue
  nodelist.append((node,bmchost)) 

if not nodelist:
    sys.stderr.write("All the requested nodes do not have an associated ipmi address.  Nothing to do.\n")
    exit(1);

if __name__ == '__main__':
    pool = multiprocessing.Pool(args.thread_count)
    results = []
    r = pool.map_async(runIpmiToolCommand, nodelist, callback=results.append)
    r.wait()
    if not results:
      sys.stderr.write("Error doing queries to nodes: %s\n" % args.moab_node_list)
      exit(1);
    print json.dumps(results[0])
