#!/usr/bin/env python

import sys, uuid, subprocess, signal, time, datetime
from optparse import OptionParser
from oslib.common import *
from acpython.mwsclient import MWSException, MWSNotFound

NODE_STATE_TIMEOUT_SECONDS = 180
SLEEP_SECONDS = 1

logger = get_logger(__name__, __file__)

option_parser = OptionParser(usage='Usage: %prog [options] <REQUESTGEOMETRY>')
option_parser.add_option('-a', '--acl', dest='acl', action='store', type='string',
    help='Set ACL on elastic nodes to REQACL', metavar='REQACL')
option_parser.add_option('-t', '--ttl-pad', dest='ttl_pad', default=180, action='store', 
    type='long', help='Pad TTL by TTLPAD seconds', metavar='TTLPAD')
option_parser.add_option('-f', '--fake-dns', dest='fake_dns', action='store_true',
    help='Workaround DNS problems by manually adding the host to the /etc/hosts file, recommended only for testing and requires the script to execute as root')


def handle_state_timeout(signum, frame):
    raise Exception('Timed out while checking if nodes were running and/or free')

def wait_for_pbs_nodes_state(server_list, state):
    signal.signal(signal.SIGALRM, handle_state_timeout)
    signal.alarm(NODE_STATE_TIMEOUT_SECONDS)
    all_are_state = False
    while not all_are_state:
        for server in server_list:
            out, err, code = call_command(['pbsnodes', '-x', server])
            if ('<state>%s</state>' % state) not in out:
                all_are_state = False
                time.sleep(SLEEP_SECONDS)
                break
            all_are_state = True
    signal.alarm(0)

def add_torque_node(name, time_to_live_datetime, request_id, acl):
    # Create node with TTL, request ID, and acl properties set
    acl_string = ''
    if acl:
        acl_string = ',acl=%s' % acl
    time_to_live_str = '%sZ' % time_to_live_datetime.replace(microsecond=0).isoformat()
    logger.debug('Adding node %s to TORQUE using TTL: %s, ACL: %s (ID: %s)' % (name, time_to_live_str, acl, request_id))
    out, err, code = call_command(['qmgr', '-c', 'create node %s TTL=%s,requestid=%s%s' % (name, time_to_live_str, request_id, acl_string)])
    if code:
        raise Exception('Could not add node %s to TORQUE (%s)' % (name, out+err))

def validate_server_count(server_count):
    valid = True
    try:
        if int(server_count) < 1:
            valid = False
    except ValueError:
        valid = False
    return valid

def get_time_to_live_datetime(time_to_live_str, ttl_pad):
    time_to_live_seconds = 0
    if ':' in time_to_live_str:
        # Parse as [[[DD:]HH:]MM:]SS
        try:
            ttl_sections = time_to_live_str.split(':')
            ttl_sections_len = len(ttl_sections)
            if ttl_sections_len > 4:
                return None
            if ttl_sections_len == 4: # days
                time_to_live_seconds += long(ttl_sections[-4]) * 86400
            if ttl_sections_len > 2: # hours
                time_to_live_seconds += long(ttl_sections[-3]) * 3600
            if ttl_sections_len > 1: # minutes
                time_to_live_seconds += long(ttl_sections[-2]) * 60
            time_to_live_seconds += long(ttl_sections[-1])
        except ValueError:
            return None
    else:
        # Parse as number of seconds
        try:
            time_to_live_seconds = long(time_to_live_str)
        except ValueError:
            return None
    if time_to_live_seconds < 1:
        return None
    time_to_live_seconds += ttl_pad
    return datetime.datetime.utcnow() + datetime.timedelta(seconds=time_to_live_seconds)

def elastic_compute(geometry, options):
    if '@' not in geometry:
        logger.error('Could not parse request geometry parameter (%s), it should be of the format <num of servers>@<time in seconds>' % geometry)
        return 1
    
    (server_count, time_to_live_str) = geometry.split('@')

    # Validate input parameters
    if not validate_server_count(server_count):
        logger.error('The server count (%s) is invalid, must be a number greater than 0' % server_count)
        return 1
    time_to_live_datetime = get_time_to_live_datetime(time_to_live_str, options.ttl_pad)
    if not time_to_live_datetime:
        logger.error('The TTL (%s) is invalid, must be a number greater than 0 or time in the format [[[DD:]HH:]MM:]SS' % time_to_live_str)
        return 1

    request_id = str(uuid.uuid1()).replace('-', '')
    logger.info('OpenStack to request %s servers with a TTL of %s UTC (ID: %s)' % (server_count, str(time_to_live_datetime), request_id))
    client = get_client(logger)
    if not client:
        return 1
    try:
        data = client.pluginService(openstack_plugin_id, 'triggerElastic',
            url_params={'serverCount': server_count, 'requestId': request_id})

        try:
            logger.info('Adding %s servers to TORQUE (ID: %s)' % (server_count, request_id))
            for server in data:
                name = server['name']
                ip_address = server['ipAddress']
            
                if options.fake_dns:
                    logger.warn('Faking DNS by adding entry for node %s to /etc/hosts, please disable this in production' % name)
                    hosts_file = open('/etc/hosts', 'a')
                    hosts_file.write('\n%s %s' % (ip_address, name))
                    hosts_file.close()

                # Add node to TORQUE
                add_torque_node(name, time_to_live_datetime, request_id, options.acl)
            node_names = [x['name'] for x in data]
            logger.debug('Waiting for %s servers to be shown as free in TORQUE (ID: %s)' % (server_count, request_id))
            wait_for_pbs_nodes_state(node_names, 'free')
            logger.info('All servers are free and ready for workload (ID: %s)' % (request_id))
            return 0
        except Exception as e:
            logger.warn(str(e))
            delete_nodes(logger, client, [x['name'] for x in data])
    except MWSNotFound:
        logger.error('A plugin named \'%s\' was not found in MWS, please see the documentation on how to create the plugin' % openstack_plugin_id)
    except MWSException as e:
        logger.error('Could not complete elastic compute request to MWS OpenStack plugin: %s' % str(e))
    return 1

def main(argv):
    options, args = option_parser.parse_args(argv)
    if len(args) != 1:
        option_parser.print_help()
        sys.exit(1)
    logger.debug('Called with parameters: %s' % ' '.join(sys.argv))
    sys.exit(elastic_compute(args[0], options))

if __name__ == '__main__':
    print 'Please see %s for logs' % logger.handlers[0].stream.name
    main(sys.argv[1:])
