import acpython
from acpython.config import getConfig
from acpython.mwsclient import MWSClient, MWSNotFound
import subprocess, logging, logging.handlers, os

openstack_plugin_id = 'open-stack'
openstack_name_pattern_config_key = 'osInstanceNamePattern'
log_level = logging.DEBUG

def get_client(logger, verbose=None):
    try:
        config = getConfig(verbose=0)
        url = config.get('mws', 'url')
        username = config.get('mws', 'username')
        password = config.get('mws', 'password')
        client = MWSClient(url=url,
                           username=username,
                           password=password,
                           verbose=0)
    except Exception as e:
        logger.error('Could not retrieve acpython configuration: %s' % str(e))
        return None
    return client

def call_command(commands):
    p = subprocess.Popen(commands, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    out, err = p.communicate()
    return (out.strip(), err.strip(), p.returncode)

def delete_nodes(logger, client, names):
    node_prefix = None
    try:
        data = client.pluginGet(openstack_plugin_id)
        # Get the string up to the first token (i.e. {date}) as the prefix
        node_prefix = data['config'][openstack_name_pattern_config_key].split('{')[0]
    except ValueError:
        logger.error('Could not find configuration parameter for \'%s\' on the \'%s\' plugin in MWS, please see documentation for correct configuration' % 
                (openstack_name_pattern_config_key, openstack_plugin_id))
    except MWSNotFound:
        logger.error('A plugin named \'%s\' was not found in MWS, please see the documentation on how to create the plugin' % openstack_plugin_id)
    except Exception as e:
        logger.warn('Could not determine the node prefix to allow deletion, ignoring deletion request: %s' % str(e))
    if not node_prefix:
        return

    for name in names:
        # Make sure the prefix matches before attempting to delete
        # This must be done in case the node idle purge time is reached for 'static' nodes that are provisioned using OpenStack
        prefix_matched = False
        try:
            if name.index(node_prefix) == 0:
                prefix_matched = True
        except ValueError:
            pass
        if not prefix_matched:
            logger.warn('Could not delete node %s, it is not a node provisioned by the OpenStack plugin (prefix: %s)' % (name, node_prefix))
            continue

        # Attempt to delete the node from OpenStack using the plugin
        try:
            logger.debug('Deleting node %s from OpenStack' % name)
            client.pluginService(openstack_plugin_id, 'triggerNodeEnd', 
                url_params={'id': name})
        except MWSNotFound:
            logger.error('A plugin named \'%s\' was not found in MWS, please see the documentation on how to create the plugin' % openstack_plugin_id)
        except Exception as e:
            logger.warn('Could not delete node %s from OpenStack %s' % (name, str(e)))

        # Attempt to delete the node from TORQUE
        try:
            logger.debug('Deleting node %s from TORQUE' % name)
            # Use call command so it consumes the standard output and error
            out, err, returncode = call_command(['qmgr', '-c', 'delete node %s' % name])
            if returncode:
                logger.warn('Could not delete node %s from TORQUE: %s' % (name, out+err))
        except Exception as e:
            logger.warn('Could not delete node %s from TORQUE: %s' % (name, str(e)))

def get_logger(name, file):
    logger = logging.getLogger(name)
    logger.setLevel(log_level)
    log_file_name = '%s.log' % os.path.splitext(os.path.basename(file))[0]
    log_file_path = os.path.join(acpython.config.getLogDir(), log_file_name)
    handler = logging.handlers.RotatingFileHandler(log_file_path, maxBytes=1024*1024*100, backupCount=3) # 100 MBs
    formatter = logging.Formatter('%(asctime)s\t%(name)s\t%(levelname)s\t%(message)s')
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    return logger
