#!/usr/bin/python
################################################################################
#
# Usage: rlm_license_check.py -f <log_file_path> [[ -d | -m | -q | -y | -v ] | [ -b <being_epoch_time> ] [ -e <end_epoch_time> ]]"
#
# This scrpt parses the RLM log file to extract the proc seconds used by dynamic nodes (used
# with Elastic Computing).
#
# Pertinent "standard" report log entries.
#
# Checkout line format:
#     OUT product version pool# user host "isv_def" count cur_use cur_resuse server_handle
#     share_handle process_id "project" "requested product" "requested version" mm/dd
#     hh:mm:ss
#
# Checkin line format:
#     IN why product version user host "isv_def" count cur_use cur_resuse server_handle
#     mm/dd hh:mm:ss
#
# AUTH section signature
#     MM/DD/YYYY HH:MM
#
#   Copyright (C) 2010-2016 by Adaptive Computing Enterprises, Inc. All Rights Reserved.
#
################################################################################

import sys, getopt
import os
import re
import time
import datetime

class rlm_log_parser(object):

    # class vars
    def __init__(self):
        self.__begin_time = ''
        self.__end_time = ''
        self.__log_file_path = ''
        self.__is_verbose = False
        self.__total_proc_seconds = 0
        self.__approx_proc_seconds = 0
        self.__day_proc_seconds = 0
        self.__month_proc_seconds = 0
        self.__quarter_proc_seconds = 0
        self.__year_proc_seconds = 0
        self.__date_pattern = '%m/%d/%Y'

    # getters/setters
    def set_begin_time(self, begin_time):
        self.__begin_time = begin_time

    def get_begin_time(self):
        return self.__begin_time

    def set_end_time(self, end_time):
        self.__end_time = end_time

    def get_end_time(self):
        return self.__end_time

    def set_log_file_path(self, log_file_path):
        self.__log_file_path = log_file_path

    def get_log_file_path(self):
        return self.__log_file_path

    def set_is_verbose(self, is_verbose):
        self.__is_verbose = is_verbose

    def get_is_verbose(self):
        return self.__is_verbose

    def get_total_proc_seconds(self):
        return self.__total_proc_seconds

    def add_to_total_proc_seconds(self, proc_seconds):
        self.__total_proc_seconds += proc_seconds 

    def add_to_day_proc_seconds(self, proc_seconds):
        self.__day_proc_seconds += proc_seconds 

    def get_day_proc_seconds(self):
        return self.__day_proc_seconds

    def get_month_proc_seconds(self):
        return self.__month_proc_seconds

    def get_quarter_proc_seconds(self):
        return self.__quarter_proc_seconds

    def get_year_proc_seconds(self):
        return self.__year_proc_seconds

    def add_to_month_proc_seconds(self, proc_seconds):
        self.__month_proc_seconds += proc_seconds 

    def add_to_quarter_proc_seconds(self, proc_seconds):
        self.__quarter_proc_seconds += proc_seconds 

    def add_to_year_proc_seconds(self, proc_seconds):
        self.__year_proc_seconds += proc_seconds 

    def set_approx_proc_seconds(self, approx_proc_seconds):
        self.__approx_proc_seconds = approx_proc_seconds 

    def get_approx_proc_seconds(self):
        return self.__approx_proc_seconds

    def get_date_pattern(self):
        return self.__date_pattern

def parse_file(rlp):

    # open the file for reading
    log_file = open(rlp.get_log_file_path(), 'r')

    auth_section_proc_seconds = 0
    last_total_licenses = 0
    last_date_time_epoch = 0
    year = 0

    day_licenses = 0
    month_licenses = 0
    quarter_licenses = 0
    year_licenses = 0

    # get begin day epoch
    this_day = datetime.datetime.now()
    begin_day_date = str(this_day.month) + '/' + str(this_day.day) + '/' + str(this_day.year)
    begin_day_epoch = int(time.mktime(time.strptime(begin_day_date, rlp.get_date_pattern())))

    # get begin month epoch
    this_month = datetime.datetime.now()
    begin_month_date = str(this_month.month) + '/1/' + str(this_month.year)
    begin_month_epoch = int(time.mktime(time.strptime(begin_month_date, rlp.get_date_pattern())))

    # get begin quarter epoch
    this_quarter = datetime.datetime.now()
    # get the first month of the quarter
    begin_quarter_month = ((this_quarter.month - 1) / 3) * 3 + 1
    begin_quarter_date = str(begin_quarter_month) + '/1/' + str(this_quarter.year)
    begin_quarter_epoch = int(time.mktime(time.strptime(begin_quarter_date, rlp.get_date_pattern())))

    # get begin year epoch
    this_year = datetime.datetime.now()
    begin_year_date = '1/1/' + str(this_year.year)
    begin_year_epoch = int(time.mktime(time.strptime(begin_year_date, rlp.get_date_pattern())))

    for log_line in log_file:

        total_licenses = 0
        date_time = ""

        # START begining section format: START aventador 08/31/2016 11:49:36
        start_match = re.search('^START\s\w+\s(\d+/\d+)/(\d+)\s(\d+:\d+:\d+)$', log_line) 
        if start_match:
            auth_section_proc_seconds = 0
            total_licenses = last_total_licenses
            date_time = start_match.group(1) + " " + start_match.group(3)
            year = start_match.group(2) 
            #import pdb; pdb.set_trace()
            if rlp.get_is_verbose() == True:
                print "Start log", start_match.group(1) + "/" + year + " " + start_match.group(3)

        # date format: 06/20/2016 13:26
        date_match = re.search('^\d+/\d+/(\d+)\s\d+:\d+$', log_line) 
        if date_match:
            if rlp.get_is_verbose() == True:
                print date_match.group(0)
            year = date_match.group(1)
            continue

        if rlp.get_is_verbose() == True:
            # AUTH format: AUTH: 1 60pg4580h028xgt1h7qubnf7qwm1n9hhp9d3cskb08a0a02rqgbu54sw4k471eamvaaw28u1sn7g
            auth_match = re.search('^AUTH:\s(\d+)\s\w+$', log_line)
            if auth_match:
                print "Auth section", auth_match.group(1), "proc seconds:", auth_section_proc_seconds
                #print auth_match.group(0)
                # reset the auth section proc seconds for the next AUTH section
                auth_section_proc_seconds = 0
                continue

        # OUT example: OUT moab_elastic_tracking 1.0 1 tshaw aventador "" 5 5 0 c3 c3 3700 "" "" "" 06/20 13:26:56
        checkout_match = re.search('^OUT\smoab_elastic_tracking\s1\.0\s\d+\s\w+\s\w+\s""\s\d+\s(\d+)\s\d+\s\w+\s\w+\s\w+\s"\w*"\s"\w*"\s"\w*"\s(.+)$', log_line)
        # IN example: IN 1 moab_elastic_tracking 1.0 tshaw aventador "" 5 10 0 c1 06/20 13:27:38 
        checkin_match = re.search('^IN\s\d+\smoab_elastic_tracking\s1\.0\s\w+\s\w+\s"\w*"\s\d+\s(\d+)\s\d+\s\w+\s(.+)$', log_line)

        #import pdb; pdb.set_trace()
        if checkout_match:
            total_licenses = checkout_match.group(1)
            date_time = checkout_match.group(2)
        elif checkin_match:
            total_licenses = checkin_match.group(1)
            date_time = checkin_match.group(2)

        if date_time:
            # convert to epoch to for easier comparison
            #import pdb; pdb.set_trace()
            time_pattern = '%m/%d/%Y %H:%M:%S'

            # add saved year from above to time extracted from log file
            pre_date_time, post_date_time = date_time.split(' ')
            new_date_time = pre_date_time + "/" + year + ' ' + post_date_time
            date_time_epoch = int(time.mktime(time.strptime(new_date_time, time_pattern)))

            # filter out data if times are given
            if rlp.get_begin_time() and date_time_epoch < int(rlp.get_begin_time()):
                continue

            if rlp.get_end_time() and date_time_epoch > int(rlp.get_end_time()):
                break

            #import pdb; pdb.set_trace()
            if last_date_time_epoch > 0:
                delta_time = date_time_epoch - last_date_time_epoch
                this_proc_seconds = int(last_total_licenses) * delta_time
                rlp.add_to_total_proc_seconds(this_proc_seconds)
                auth_section_proc_seconds += this_proc_seconds

                # set daily
                if date_time_epoch >= begin_day_epoch:
                    rlp.add_to_day_proc_seconds(this_proc_seconds)

                # set monthly 
                if date_time_epoch >= begin_month_epoch:
                    rlp.add_to_month_proc_seconds(this_proc_seconds)

                # set quarterly 
                if date_time_epoch >= begin_quarter_epoch:
                    rlp.add_to_quarter_proc_seconds(this_proc_seconds)

                # set yearly
                if date_time_epoch >= begin_year_epoch:
                    rlp.add_to_year_proc_seconds(this_proc_seconds)

            # save off time and licenses
            last_date_time_epoch = date_time_epoch
            last_total_licenses = total_licenses
          
    log_file.close()

    # We're done parsing the file. If all license aren't checked in, calculate proc second usage 
    # of those licenses up to this point

    #import pdb; pdb.set_trace()
    if int(last_total_licenses) is not 0 and last_date_time_epoch is not 0:
        date_time_epoch = int(time.time())    
        delta_time = date_time_epoch - last_date_time_epoch
        this_proc_seconds = int(last_total_licenses) * delta_time
        rlp.set_approx_proc_seconds(rlp.get_total_proc_seconds() + this_proc_seconds)


def print_usage():

    print "Usage:    rlm_parse_log_file.py -f <log_file_path> [[ -v ] | [ -b <being_epoch_time> ] [ -e <end_epoch_time> ]]"


def parse_cmd_line_args(argv, rlp):

    try:
        opts, args = getopt.getopt(argv,'b:de:f:hmqvy')
    except getopt.GetoptError:
        print_usage()
        sys.exit(2)

    for opt, arg in opts:
        if opt == '-b':
            rlp.set_begin_time(arg)
        elif opt == '-e':
            rlp.set_end_time(arg)
        elif opt == '-f':
            rlp.set_log_file_path(arg)
        elif opt == '-h':
            print_usage()
            sys.exit(0)
        elif opt == '-v':
            rlp.set_is_verbose(True)


def main(argv):

    rlp = rlm_log_parser()

    parse_cmd_line_args(argv, rlp)

    # check to make sure the file is a regular file

    if not os.path.isfile(rlp.get_log_file_path()):
        print "Invalid or nonexistent log file"
        print_usage()
        sys.exit(2)

    parse_file(rlp)

    print "This day =", rlp.get_day_proc_seconds()
    print "This month =", rlp.get_month_proc_seconds()
    print "This quarter =", rlp.get_quarter_proc_seconds()
    print "This year =", rlp.get_year_proc_seconds()

    if rlp.get_approx_proc_seconds() is not 0:
        print "Total approx current proc seconds ~", rlp.get_approx_proc_seconds()

    print "Total finished proc seconds =", rlp.get_total_proc_seconds()


if __name__ == '__main__':
    main(sys.argv[1:])
