#!/usr/bin/python

################################################################################
#
# Usage: sync_moab_rlm.py [[ -h <host> | -l <rlm_report_log> | -s <rlm_parse_log_file> | -u <username> ]]
#
# NOTE: hostname and username only necessary if rlm server is on remote host
#
# This script runs the rlm_parse_log_file.py script and uses the output to synchronize
# the proc seconds with Moab
#
#   Copyright (C) 2010-2016 by Adaptive Computing Enterprises, Inc. All Rights Reserved.
#
################################################################################


import sys, getopt
import os
import re
import subprocess

class rlm_remote_connection(object):

    def __init__(self):
        self.__is_remote = False
        self.__report_log_file = ''
        self.__remote_host = ''
        self.__rlm_parse_script = ''
        self.__remote_user = ''

    def is_remote(self):
        return self.__is_remote

    def set_report_log_file(self, report_log_file):
        self.__report_log_file = report_log_file 

    def get_report_log_file(self):
        return self.__report_log_file

    def set_remote_host(self, host):
        self.__remote_host = host
        self.__is_remote = True

    def get_remote_host(self):
        return self.__remote_host

    def set_rlm_parse_script(self, rlm_parse_script):
        self.__rlm_parse_script = rlm_parse_script

    def get_rlm_parse_script(self):
        return self.__rlm_parse_script

    def set_remote_user(self, remote_user):
        self.__remote_user = remote_user 

    def get_remote_user(self):
        return self.__remote_user



def parse_cmd_line_args(argv, rrc):

    try:
        opts, args = getopt.getopt(argv,'f:h:s:u:')
    except getopt.GetoptError:
        print_usage()
        sys.exit(2)

    for opt, arg in opts:
        if opt == '-f':
            rrc.set_report_log_file(arg)
        elif opt == '-h':
            rrc.set_remote_host(arg)
        elif opt == '-s':
            rrc.set_rlm_parse_script(arg)
        elif opt == '-u':
            rrc.set_remote_user(arg)


def print_usage():

    print "Usage:    sync_moab_rlm.py [[ -f <rlm_report_log> | -h <host> | -s <rlm_parse_log_file> | -u <username> ]]"


def main(argv):

    #import pdb; pdb.set_trace()
    rrc = rlm_remote_connection()
    
    parse_cmd_line_args(argv, rrc)

    moab_home_dir = os.environ['MOABHOMEDIR']

    if not moab_home_dir:
        moab_home_dir = "/opt/moab"

    rlm_parse_script = rrc.get_rlm_parse_script()

    if not rlm_parse_script:
        rlm_parse_script = moab_home_dir + "/tools/rlm/rlm_parse_log_file.py"

    report_log_file = rrc.get_report_log_file()

    if not report_log_file:
        report_log_file = moab_home_dir + "/log/adaptiveco.log"

    if rrc.is_remote() == True:
        rlm_parse_script_cmd = "ssh " + rrc.get_remote_host()
        remote_user = rrc.get_remote_user()
        if remote_user:
            rlm_parse_script_cmd += " -l " + remote_user
        rlm_parse_script_cmd += " \"" + rlm_parse_script + " -f " + report_log_file + "\""

        #print rlm_parse_script_cmd

        # ssh to remote server and run command
        try:
            proc = subprocess.Popen(rlm_parse_script_cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
            (output, err) = proc.communicate()
            #print output
        except subprocess.CalledProcessError:
            print "Error running rlm_parse_cmd", err
            sys.exit(2)
    else:
        #print rlm_parse_script, "-f", report_log_file 
        try:
            proc = subprocess.Popen([rlm_parse_script, "-f", report_log_file], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            (output, err) = proc.communicate()
            #print output
        except subprocess.CalledProcessError:
            print "Error running rlm_parse_cmd", err
            sys.exit(2)

    day_proc_seconds = 0
    month_proc_seconds = 0
    quarter_proc_seconds = 0
    year_proc_seconds = 0
    time_frame_list = []

    day_match = re.search('This\sday\s=\s(\d+)?', output)
    month_match = re.search('This\smonth\s=\s(\d+)?', output)
    quarter_match = re.search('This\squarter\s=\s(\d+)?', output)
    year_match = re.search('This\syear\s=\s(\d+)?', output)

    if day_match:
        time_frame_list.append("daily=" + day_match.group(1))
    else:
        print "Error parsing last day output from rlm_parse_cmd.py", err
        sys.exit(2)

    if month_match:
        time_frame_list.append("monthly=" + month_match.group(1))
    else:
        print "Error parsing last month output from rlm_parse_cmd.py", err
        sys.exit(2)

    if quarter_match:
        time_frame_list.append("quarterly=" + quarter_match.group(1))
    else:
        print "Error parsing last quarter output from rlm_parse_cmd.py", err
        sys.exit(2)

    if year_match:
        time_frame_list.append("yearly=" + year_match.group(1))
    else:
        print "Error parsing last year output from rlm_parse_cmd.py", err
        sys.exit(2)


    mschedctl_cmd = moab_home_dir + "/bin/mschedctl"

    for time_frame in time_frame_list:
        try:
            proc = subprocess.Popen([mschedctl_cmd, "-m", "procseconds", time_frame], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            (output, err) = proc.communicate()
            #print output
            #print err 
        except subprocess.CalledProcessError:
            print "Error running mschedctl command", err
            sys.exit(2)


if __name__ == '__main__':
    main(sys.argv[1:])

